#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Purpose
-------

This module allows authorized users to upload chewBBACA's schemas
to the Chewie-NS.

The process for schema upload has four stages:

    - User Permissions: Determines if the current user has permission
      to upload schemas. Only Admin or Contributor level users can
      upload schemas to the Chewie-NS.

    - Parameters Validation: Validation of the set of parameters associated
      with the schema. Only schemas that have been used with a single valid
      value per parameter can be uploaded. Invalid or multiple values
      for a single parameter can lead to inconsistent results; thus,
      it is strongly advised to always perform allele calling with
      the same set of parameters and refrain from altering the initial
      set of parameters values defined in the schema creation or
      adaptation processes.

    - Schema Pre-processing: Applies quality control measures to identify
      and exclude invalid alleles. Searches for annotations on UniProt
      and imports annotations provided by users.

    - Schema Upload: Collects essential data and sends it to the Chewie-NS
      for schema creation and data insertion. The process finishes when all
      the necessary data has been uploaded. The Chewie-NS automatically
      detects that all data has been received and finishes data insertion.

Expected input
--------------

The process expects the following variables whether through command line
execution or invocation of the :py:func:`main` function:

- ``-i``, ``schema_directory`` : Path to the directory of the schema to upload.

    - e.g.: ``/home/user/schemas/ypestis_schema``

- ``-sp``, ``species_id`` : The integer identifier or name of the species that
  the schema will be associated to in the Chewie-NS.

    - e.g.: ``1`` or ``'Yersinia pestis'``

- ``-sn``, ``schema_name`` : A brief and meaningful name that should help
  understand the type and content of the schema.

    - e.g.: ``ypestis_cgMLST`` or ``ypestis cgMLST``

- ``-lp``, ``loci_prefix`` : Prefix included in the name of each locus of the
  schema.

    - e.g.: ``ypestis``

- ``--df``, ``description_file`` : Path to a text file with a description
  about the schema. Markdown syntax is supported in order to allow greater
  customizability of the rendered description in the Frontend. Will default
  to the schema's name if the user does not provide a valid path for a
  file (default=None).

    - e.g.: ``/home/user/schemas/ypestis_description``

- ``--a``, ``annotations`` : Path to a TSV file with loci annotations. The
  first column has loci identifiers (w/o .fasta extension), the second has
  user annotations and the third has custom annotations (default=None).

    - e.g.: ``/home/user/schemas/ypestis_annotations``

- ``--cpu``, ``cpu_cores`` : Number of CPU cores that will be used in the
  Schema Pre-processing step (default=1).

    - e.g.: ``4``

- ``--ns_url``, ``nomenclature_server`` : The base URL for the Nomenclature
  Server. The default value, "main", will establish a connection to
  "https://chewbbaca.online/", "tutorial" to "https://tutorial.chewbbaca.online/"
  and "local" to "http://127.0.0.1:5000/NS/api/" (localhost). Users may also
  provide the IP address to other Chewie-NS instances.

    - e.g.: ``http://127.0.0.1:5000/NS/api/`` (localhost)

- ``--continue_up`` : If the process should check if the schema upload was
  interrupted and try to resume it. ``True`` if provided, ``False`` otherwise.

Code documentation
------------------
"""


import os
import sys
import json
import time
import hashlib
import argparse
import requests
import itertools
import multiprocessing
import concurrent.futures

from Bio import SeqIO
from SPARQLWrapper import SPARQLWrapper, JSON
from urllib3.exceptions import InsecureRequestWarning

try:
    from utils import (constants as ct,
                       file_operations as fo,
                       uniprot_requests as ur,
                       chewiens_requests as cr,
                       fasta_operations as fao,
                       parameters_validation as pv,
                       sequence_manipulation as sm)
except ModuleNotFoundError:
    from CHEWBBACA.utils import (constants as ct,
                                 file_operations as fo,
                                 uniprot_requests as ur,
                                 chewiens_requests as cr,
                                 fasta_operations as fao,
                                 parameters_validation as pv,
                                 sequence_manipulation as sm)


# Suppress only the single warning from urllib3 needed.
requests.packages.urllib3.disable_warnings(category=InsecureRequestWarning)


uniprot_sparql = SPARQLWrapper(ct.UNIPROT_SPARQL)


def import_annotations(tsv_file):
    """Read a TSV file to get loci annotations.

    This function imports loci annotations from
    a user-provided TSV file. Users can provide
    annotations for the full set of loci in the
    schema or only for a subset.

    Parameters
    ----------
    tsv_file : str
        Path to a TSV file with loci annotations.
        The TSV file has the following three columns:

        - Loci identifiers (schema's filenames stripped of
          the '.fasta' suffix).
        - User annotation, a term that the user attributes
          to each locus.
        - Custom annotation, an additional term that the
          user might want to provide.

        It is not mandatory to provide terms for both
        annotation fields.

    Returns
    -------
    annotations : dict
        Dictionary with loci identifiers as keys and lists
        with user and custom annotations as values.
    """
    # read file
    lines = fo.read_tabular(tsv_file)

    # fill missing values with N/A
    lines = [line + ['N/A']*(3-len(line)) for line in lines]

    annotations = {}
    for line in lines:
        locus = line[0]
        # do not keep blank values
        a = ['N/A' if i.strip() == '' else i for i in line[1:]]

        # only keep cases that have at least one non-NA field
        if a[0] != 'N/A' or a[1] != 'N/A':
            annotations[locus] = a

    return annotations


def search_schema(schemas, schema_name):
    """Determine if any of the schemas in the input list has the provided name.

    Parameters
    ----------
    schemas : list of dict
        A list that contains dictionaries.
        Each dictionary has information about a schema.
    schema_name : str
        Name of the schema to look for in the input list
        of schemas.

    Returns
    -------
    list
        A list with the schema URI and schema identifier
        if the schema is found or a list with the boolean
        value of False if it is not found.
    """
    match = [s for s in schemas if s['name']['value'] == schema_name]

    if len(match) > 0:
        schema = match[0]
        schema_uri = schema['schemas']['value']
        schema_id = schema_uri.split('/')[-1]

        return [schema_uri, schema_id]
    else:
        return [False]


def schema_status(base_url, headers_get, schema_name, species_id, continue_up):
    """Determine if it is possible to create a schema or resume its upload.

    Parameters
    ----------
    base_url : str
        Base URL of the Chewie Nomenclature Server.
    headers_get : dict
        HTTP headers for GET requests.
    schema_name : str
        Name of the schema that the user wants to create
        or resume the upload of.
    species_id : str
        The identifier of the schema's species in the
        Chewie-NS.
    continue_up : bool
        False if the schema is to be uploaded as a new schema
        or True if the schema already exists.

    Returns
    -------
    list
        A list with the upload type ('novel' for new schemas,
        'incomplete' for upload resuming) and the schema
        identifier (NoneType if the schema still does not
        exist).

    Raises
    ------
    SystemExit
        - If `continue_up` is True and the schema does not exist.
        - If `continue_up` is False and a schema with name value
          equal to `schema_name` already exists.
        - If `continue_up` is True and the schema has been fully
          uploaded.
        - If `continue_up` is True and the current user is not the
          one that administrates the schema.
    """
    schema_id = None
    schemas_url, schemas_get = cr.simple_get_request(base_url, headers_get,
                                                     ['species', species_id, 'schemas'])
    schemas_status = schemas_get.status_code
    if schemas_status in [200, 201]:
        species_schemas = schemas_get.json()
        # determine if there is a schema for current
        # species with provided name
        schema_info = search_schema(species_schemas, schema_name)

        # there is no schema with provided name
        if schema_info[0] is False:
            if continue_up is False:
                upload_type = 'novel'
            elif continue_up is True:
                sys.exit('Cannot continue uploading a schema that '
                         'does not exist.')
        # there is a schema with provided name
        else:
            if continue_up is False:
                sys.exit('Cannot upload schema. A schema with '
                         'provided name already exists.')
            elif continue_up is True:
                schema_url, schema_id = schema_info
                # determine if schema upload was complete
                schema_get = cr.simple_get_request(base_url, headers_get,
                                                   ['species', species_id,
                                                    'schemas', schema_id])[1]
                current_schema = schema_get.json()[0]
                schema_date = current_schema['dateEntered']['value']
                if schema_date != 'singularity':
                    sys.exit('Schema finished uploading. Cannot proceed.')
                # determine if user was the one that started the upload
                ask_admin = cr.simple_get_request(base_url, headers_get,
                                                  ['species', species_id,
                                                   'schemas', schema_id,
                                                   'administrated'])[1]
                schema_admin = ask_admin.json()
                if schema_admin is False:
                    sys.exit('Current user is not the user that '
                             'started schema upload. Exited.')
                upload_type = 'incomplete'
    else:
        if continue_up is True:
            sys.exit('Cannot continue uploading a schema that '
                     'does not exist.')
        else:
            upload_type = 'novel'

    return [upload_type, schema_id]


def create_uniprot_queries(file, max_queries=ct.MAX_QUERIES):
    """Create SPARQL queries to search for protein annotations on UniProt.

    Parameters
    ----------
    file : str
        Path to a pickled file that contains a dictionary
        with protein identifiers as keys and protein
        sequences as values.
    max_queries : int
        Maximum number of queries that will be created.
        Only the first `max_queries` unique proteins will
        be used to create queries.

    Returns
    -------
    queries_file : str
        Path to the pickled file that contains a list
        with the SPARQL queries that were created.
    """
    # import dictionary with protein sequences as values
    protein_seqs = fo.pickle_loader(file)

    # create queries based on unique sequences only
    unique_prots = set(list(protein_seqs.values()))
    selected = unique_prots if len(unique_prots) <= max_queries \
        else list(unique_prots)[0:max_queries]

    queries = [ur.uniprot_query(prot) for prot in selected]
    # save SPARQL queries with pickle
    queries_file = '{0}_up'.format(file)
    fo.pickle_dumper(queries, queries_file)

    return queries_file


def get_annotation(queries_file, max_queries=ct.MAX_QUERIES):
    """Query the UniProt SPARQL endpoint to retrieve protein annotations.

        This function uses the SPARQL queries imported from
        a pickled file to search for protein annotations
        through UniProt's SPARQL endpoint. It searches until
        it retrieves terms that are considered informative or
        until it reaches the maximum number of searches that is
        allowed.

    Parameters
    ----------
    queries_file : str
        Path to a pickled file that contains a list with
        SPARQL queries that can be used to search for protein
        annotations.

    Returns
    -------
    annotation_info : list of str
        A list with the following elements:

        - Locus identifier.
        - UniProt annotation name.
        - UniProt annotation label.
        - UniProt annotation URL.
    """
    locus = queries_file.split('/')[-1].split('_prots_up')[0]

    # load queries for locus
    queries = fo.pickle_loader(queries_file)

    uniprot_sparql.setReturnFormat(JSON)
    uniprot_sparql.setTimeout(10)

    prev_url = 'N/A'
    prev_name = 'N/A'
    prev_label = 'N/A'
    uninformative = ['Uncharacterized protein',
                     'hypothetical protein',
                     'DUF']

    tries = 0
    # define maximum number of tries
    max_tries = max_queries
    found = False
    while found is False:

        uniprot_sparql.setQuery(queries[tries])

        try:
            result = uniprot_sparql.query().convert()
            name, url, label = ur.select_name(result)

            if name != '':
                if prev_name == 'N/A':
                    prev_name = name
                    prev_label = label
                    prev_url = url
                # keep latest term if previous term was uninformative
                elif any([n in prev_name for n in uninformative]) is True \
                        and any([n in name for n in uninformative]) is False:
                    prev_name = name
                    prev_label = label
                    prev_url = url

            if prev_name not in uninformative:
                found = True

        except Exception:
            pass

        tries += 1
        if tries == max_tries or tries == len(queries):
            found = True

    annotation_info = [locus, prev_name, prev_label, prev_url]

    return annotation_info


def quality_control(locus_input):
    """Translate DNA sequences and report invalid alleles.

    Parameters
    ----------
    locus_input : list
        A list with the following elements:

        - Path to the FASTA file with the locus DNA
          sequences (str).
        - Locus identifier (str).
        - Genetic code that should be used to translate
          the DNA sequences (int).
        - Minimum sequence length value (int). Alleles
          with a length value that is lower than this value
          will be considered invalid.
        - Sequence size threshold value (NoneType). Set
          to None to avoid removal of alleles that deviate
          from the sequence length mode based on the current
          set of sequences, but that did not deviate when they
          were added.

    Returns
    -------
    list
        A list with the following elements:

        - Path to the FASTA file with the locus DNA
          sequences (str).
        - Path to a pickled file that contains a dictionary
          with protein identifiers as keys and protein sequences
          as values.
        - A list with the identifiers of invalid alleles
          (list of str).
    """
    res = sm.get_seqs_dicts(*locus_input)

    prots_file = '{0}_prots'.format(locus_input[0].split('.fasta')[0])
    fo.pickle_dumper(res[1], prots_file)

    if len(res[2]) > 0:
        print('  Found {0} invalid alleles for '
              'locus {1}.'.format(len(res[2]), locus_input[1]))

    return [locus_input[0], prots_file, res[2]]


def create_lengths_files(loci_files, out_dir):
    """Determine the length of the sequences in a set of FASTA files.

    Parameters
    ----------
    loci_files : list of str
        List with paths to FASTA files.
    out_dir : str
        Directory where the output files with the length
        values of the sequences from each input FASTA file
        will be created.

    Returns
    -------
    length_files : list of str
        List with paths to the pickled files that contain
        the sequence length values determined for each input
        FASTA file. Each pickled file contains a dictionary
        with hashes (sequence hashes computed with sha256)
        as keys and sequence lengths as values.
    """
    length_files = []
    for file in loci_files:
        locus_basename = fo.file_basename(file)
        locus_lengths = {locus_basename: fao.sequence_lengths(file, True)}
        lengths_file = os.path.join(out_dir,
                                    '{0}_lengths'.format(locus_basename.split('.fasta')[0]))

        fo.pickle_dumper(locus_lengths, lengths_file)
        length_files.append(lengths_file)

    return length_files


def schema_completedness(base_url, species_id, schema_id, headers_get,
                         hashed_files):
    """Determine a schema's upload status.

    Returns information that indicates which schema data has
    been or hasn't been fully uploaded and inserted into the
    Chewie-NS database.

    Parameters
    ----------
    base_url : str
        Base URL of the Chewie Nomenclature server.
    species_id : int
        The identifier of the schema's species in the NS.
    schema_id : int
        The identifier of the schema in the NS.
    headers_get : dict
        HTTP headers for GET requests.
    hashed_files : dict
        Dictionary with file hashes (obtained through BLAKE2
        hashing of file content) as keys and file paths as
        values.

    Returns
    -------
    list
        A list with the following elements:

        - A dictionary with paths to loci FASTA files as keys
          and boolean values that indicate if loci data has been
          uploaded and inserted.
        - A dictionary with the same structure as the previous
          one but that only contains entries for loci that were
          not fully inserted.
        - A list with the paths to FASTA files of the loci whose
          set of alleles was not fully uploaded.

        The first and second elements, both dictionaries, have
        lists with four elements as values. The first, second
        and third elements on those lists are Boolean values
        that indicate if the locus has been created, linked to
        the species and linked to the schema, respectively. The
        last element is the BLAKE2 hash of the locus file.
    """
    # get info about loci and alleles upload
    schema_loci = cr.simple_get_request(base_url, headers_get,
                                        ['species', species_id,
                                         'schemas', schema_id,
                                         'loci', 'data'])[1]
    schema_loci = schema_loci.json()
    if 'Not found' in schema_loci:
        sys.exit('{0}'.format(schema_loci['Not found']))
    else:
        schema_loci = schema_loci['hashes']

    loci_info = {hashed_files[k]: v[1]+[k] for k, v in schema_loci.items()}
    # determine loci that were not fully inserted
    absent_loci = {hashed_files[k]: v[1]+[k]
                   for k, v in schema_loci.items() if all(v[1]) is not True}
    # determine loci without alleles
    absent_alleles = [hashed_files[k]
                      for k, v in schema_loci.items() if v[0] is False]

    return [loci_info, absent_loci, absent_alleles]


def create_schema(base_url, headers_post, species_id, params):
    """Create a schema in Chewie-NS.

    Parameters
    ----------
    base_url : str
        Base URL of the Nomenclature server.
    headers_post : dict
        HTTP headers for POST requests.
    species_id : int
        The identifier of the schema's species in the
        Chewie-NS.
    params : dict
        A dictionary with the following key/values pairs:

        - bsr (str): BLAST Score Ratio value used to create the
          schema and perform allele calling.
        - prodigal_training_file (str): BLAKE2 hash of the Prodigal
          training file content.
        - translation_table (str): genetic code used to predict
          and translate coding sequences.
        - minimum_locus_length (str): minimum sequence length,
          sequences with a length value lower than this value
          are not included in the schema.
        - chewBBACA_version (str): version of the chewBBACA suite
          used to create the schema and perform allele calling.
        - size_threshold (str): sequence size variation percentage
          threshold, new alleles cannot have a length value that
          deviates +/- than this value in relation to the locus's
          representative sequence.
        - word_size (str): word/k value used to cluster protein
          sequences during schema creation and allele calling.
        - cluster_sim (str): proportion of k-mers that two proteins
          need to have in common to be clustered together.
        - representative_filter (str): proportion of k-mers that a
          clustered protein has to have in common with the representative
          protein of the cluster to be considered the same gene.
        - intraCluster_filter (str): proportion of k-mers that clustered
          proteins have to have in common to be considered of the same
          gene.
        - name (str): name of the new schema. Must be unique among
          the schemas' of the species.
        - SchemaDescription (str): BLAKE2 of the contents of the file
          with the schema description.
        - schema_hashes (list of str): list that contains the BLAKE2
          hashes of the schema files.

    Returns
    -------
    list
        A list with the following elements: the schema URI
        (str) and the schema identifier (str) in the Chewie-NS.
    """
    schema_post = cr.simple_post_request(base_url, headers_post,
                                         ['species', species_id, 'schemas'],
                                         data=json.dumps(params))[1]
    schema_status = schema_post.status_code

    # check status code
    if schema_status not in [200, 201]:
        print(schema_status, schema_post.text)
        sys.exit('Could not create new schema in the NS.')

    schema_url = schema_post.json()['url']
    schema_id = schema_url.split('/')[-1]

    return [schema_url, schema_id]


def create_loci_file(schema_files, annotations, schema_dir,
                     species_id, schema_id, loci_prefix,
                     absent_loci=None):
    """Create a file with the essential data to insert loci in Chewie-NS.

    Parameters
    ----------
    schema_files : list of str
        List with paths to schema's FASTA files.
    annotations : dict
        Dictionary with loci identifiers as keys and lists
        with annotation terms as values.
    schema_dir : str
        Path to the directory with schema files.
    species_id : int
        The identifier of the schema's species in the NS.
    schema_id : int
        The identifier of the schema in the NS.
    loci_prefix : str
        Prefix to include in loci identifiers.
    absent_loci : dict, optional
        Loci that have not been created or are incomplete.
        Default value of None collects and saves data
        from all schema's loci.

    Returns
    -------
    loci_file : str
        Path to the file with the necessary data to insert loci
        and associate with species and schema.
    """
    loci_file = os.path.join(schema_dir,
                             '{0}_{1}_loci'.format(species_id, schema_id))
    loci_data = [loci_prefix, []]
    for file in schema_files:
        if absent_loci is None or file in absent_loci:

            locus = os.path.basename(file).split('.fasta')[0]
            locus_hash = fo.hash_file(file, hashlib.blake2b())

            locus_annotations = annotations[locus]

            loci_data[1].append([locus, locus_hash]+locus_annotations)

    if len(loci_data[1]) > 0:
        fo.pickle_dumper(loci_data, loci_file)

    return loci_file


def upload_loci_data(loci_file, base_url, species_id,
                     schema_id, headers_post, headers_get, hashed_files):
    """Send data to insert and associate loci to a species and schema.

    Parameters
    ----------
    loci_file : str
        Path to the file with the data to upload to the
        Chewie-NS.
    base_url : str
        Base URL of the Nomenclature server.
    species_id : int
        The identifier of the schema's species in the NS.
    schema_id : int
        The identifier of the schema in the NS.
    headers_post : dict
        HTTP headers for POST requests.
    hashed_files : dict
        Dictionary with file hashes as keys and file paths
        as values.

    Returns
    -------
    response_data : dict
        Dictionary with paths to loci FASTA files as keys and
        lists with four elements as values. The first, second
        and third elements on those lists are Boolean values
        that indicate if the locus has been created, linked to
        the species and linked to the schema, respectively. The
        last element is the BLAKE2 hash of the locus file.

    Raises
    ------
    SystemExit

        - If the data sent to the NS does not match the data that
          was determined to be in the schema.
        - If the response returned from the NS indicates that it
          was not possible to create and associate any/some loci.
    """
    # compress file with loci data to reduce upload size
    loci_zip_file = '{0}.zip'.format(loci_file)
    fo.file_zipper(loci_file, loci_zip_file)

    zip_url = cr.make_url(base_url, 'species', species_id,
                          'schemas', schema_id, 'loci',
                          'data')

    # upload file as multipart-encoded file
    filename = '{0}_{1}_loci.zip'.format(species_id, schema_id)
    response = cr.upload_file(loci_zip_file, filename,
                              zip_url, headers_post,
                              False)

    response = response.json()
    start_nr_loci = int(response['nr_loci'])
    start_sp_loci = int(response['sp_loci'])
    start_sc_loci = int(response['sc_loci'])

    time_limit = 2100
    current_time = 0
    status = 'Inserting'
    while status != 'Complete' and (current_time < time_limit):
        insertion_status = cr.simple_get_request(base_url, headers_get,
                                                 ['species', species_id,
                                                  'schemas', schema_id,
                                                  'loci', 'data'])[1]
        insertion_status = insertion_status.json()

        if insertion_status['status'] == 'complete':
            status = 'Complete'
            response_data = insertion_status['hashes']

        nr_loci = int(insertion_status['nr_loci'])
        sp_loci = int(insertion_status['sp_loci'])
        sc_loci = int(insertion_status['sc_loci'])

        inserted_loci = nr_loci - start_nr_loci
        linked_sp = sp_loci - start_sp_loci
        linked_sc = sc_loci - start_sc_loci

        print('\r', '   Inserted {0} loci; '
              'Linked {1} to species; '
              'Linked {2} to schema.'.format(inserted_loci, linked_sp, linked_sc), end='')
        time.sleep(2)
        current_time += 2

    response_data = {hashed_files[k]: v+[k] for k, v in response_data.items()}

    return response_data


def create_alleles_files(schema_files, loci_responses, invalid_alleles,
                         species_name, base_url, species_id,
                         schema_id, user_id):
    """Create files with data to insert alleles in Chewie-NS.

    Parameters
    ----------
    schema_files : list
        List with paths to schema's FASTA files.
    loci_responses : dict
        Dictionary with files paths as keys and
    invalid_alleles : list
        List with the identifiers of the alleles
        that were determined to be invalid.
    species_name : str
        Name of the species.
    base_url : str
        Base URL of the Nomenclature server.
    species_id : int
        The identifier of the schema's species in the NS.
    schema_id : int
        The identifier of the schema in the NS.
    user_id : int
        Current user identifier.

    Returns
    -------
    list of list
        List with the following elements:

        - List with paths to files that contain the data that
          is necessary to upload to insert all alleles.
        - List with the loci identifiers in the NS.
        - List with file hashes for loci files.
        - List with loci basenames.
    """
    loci_ids = []
    loci_names = []
    loci_hashes = []
    alleles_files = []
    schema_dir = os.path.dirname(schema_files[0])
    user_uri = '{0}users/{1}'.format(base_url, user_id)
    for file in schema_files:
        locus_uri = loci_responses[file][1][0]

        alleles_sequences = [str(rec.seq)
                             for rec in SeqIO.parse(file, 'fasta')
                             if rec.id not in invalid_alleles]

        post_inputs = [locus_uri, species_name,
                       user_uri, tuple(alleles_sequences)]

        locus_id = locus_uri.split('/')[-1]
        loci_ids.append(locus_id)

        locus_hash = loci_responses[file][-1]
        loci_hashes.append(locus_hash)

        locus_name = os.path.basename(file).split('.fasta')[0]
        loci_names.append(locus_name)

        alleles_file = os.path.join(schema_dir,
                                    '{0}_{1}_{2}'.format(species_id,
                                                         schema_id,
                                                         locus_id))
        alleles_files.append(alleles_file)

        fo.pickle_dumper(post_inputs, alleles_file)

    return [alleles_files, loci_ids, loci_hashes, loci_names]


def upload_alleles_data(alleles_data, length_files, base_url,
                        headers_post, headers_post_bytes,
                        species_id, schema_id):
    """Upload files with data to insert alleles.

    Parameters
    ----------
    alleles_data : list
        List with tuples, one per locus, that contain the path
        to the ZIP archive with the data to insert alleles,
        the identifier of the locus, the locus file hash and
        the basename of the locus file.
    length_files : list
        List with paths to the pickled files that contain a
        dictionary with sequences hashes as keys and sequence
        length as values.
    base_url : str
        Base URL of the Nomenclature server.
    headers_post : dict
        HTTP headers for POST requests that accept JSON
        formatted data.
    headers_post_bytes : dict
        HTTP headers for POST requests that support file
        upload.
    species_id : int
        The identifier of the schema's species in the NS.
    schema_id : int
        The identifier of the schema in the NS.

    Returns
    -------
    failed : list of str
        List with the identifiers of the loci whose alleles
        data could not be fully uploaded.
    """
    failed = []
    uploaded = 0
    for i, a in enumerate(alleles_data):

        current_locus = a[1]
        locus_hash = a[2]

        # get length of alleles from current locus
        current_len = length_files[i]
        data = fo.pickle_loader(current_len)
        data = {current_locus: data[next(iter(data))]}
        data = {'content': data, 'locus_hash': locus_hash}

        # send data to the NS
        send_url = cr.make_url(base_url, 'species', species_id,
                               'schemas', schema_id, 'loci',
                               current_locus, 'lengths')

        lengths_res = cr.simple_post_request(send_url, headers_post,
                                             data=json.dumps(data))
        length_status = lengths_res[1].status_code

        # get path to ZIP archive with data to insert alleles
        current_zip = a[0]

        # send data to insert alleles in the NS
        zip_url = cr.make_url(base_url, 'species', species_id,
                              'schemas', schema_id, 'loci',
                              current_locus, 'data')

        zip_res = cr.upload_file(current_zip, locus_hash,
                                 zip_url, headers_post_bytes,
                                 False)
        zip_status = zip_res.status_code

        # determine if upload was successful
        if length_status not in [200, 201] or zip_status not in [200, 201]:
            failed.append(current_locus)
        elif length_status in [200, 201] and zip_status in [200, 201]:
            uploaded += 1
            print('\r', '    Sent data for alleles of '
                  '{0} loci.'.format(uploaded), end='')

    return failed


def main(schema_directory, species_id, schema_name, loci_prefix,
         description_file, annotations, cpu_cores,
         nomenclature_server, continue_up):

    if 'tutorial' not in nomenclature_server:
        token = cr.capture_login_credentials(nomenclature_server)
    else:
        token = ''

    # verify user
    print('-- User Permissions --')
    # GET request headers
    headers_get = ct.HEADERS_GET_JSON
    headers_get['Authorization'] = token

    # determine current user ID and Role
    if 'tutorial' not in nomenclature_server:
        user_id, user_role, user_auth = cr.user_info(nomenclature_server, headers_get)
    else:
        user_id, user_role, user_auth = ['', '', True]

    print('User id: {0}'.format(user_id))
    print('User role: {0}'.format(user_role))
    print('Authorized: {0}\n'.format(user_auth))

    # only Admin or Contributor type users can upload schemas
    if not user_auth:
        sys.exit('Current user has no Administrator '
                 'or Contributor permissions.\n'
                 'Not allowed to upload schemas.')

    # POST requests headers
    headers_post = ct.HEADERS_POST_JSON
    headers_post['Authorization'] = token
    headers_post['user_id'] = user_id
    # POST headers to send binary data
    headers_post_bytes = ct.HEADERS_POST
    headers_post_bytes['Authorization'] = token
    headers_post_bytes['user_id'] = user_id

    print('-- Parameters Validation --')
    print('Local schema: {0}'.format(schema_directory))

    # Get schema files from genes list file
    genes_list = os.path.join(schema_directory, '.genes_list')
    genes = fo.pickle_loader(genes_list)

    fasta_paths = [os.path.join(schema_directory, file) for file in genes]
    fasta_paths.sort()

    # total number of loci
    total_loci = len(fasta_paths)
    # total number of alelles
    total_alleles = sum(list(map(fao.count_sequences, fasta_paths)))

    # Get the name of the species from the provided id
    # or vice-versa
    species_info = cr.species_ids(species_id, nomenclature_server, headers_get)
    if isinstance(species_info, list):
        species_id, species_name = species_info
        print("Schema's species: {0} (id={1})".format(species_name, species_id))
    else:
        sys.exit('There is no species with the provided identifier in the NS.')

    print('Number of loci: {0}'.format(total_loci))
    print('Number of alleles: {0}\n'.format(total_alleles))

    # verify schema configs
    print('Verifying schema configs...')
    # load schema config file
    configs = pv.read_configs(schema_directory, '.schema_config')

    # validate arguments values
    schema_ptfs = configs['prodigal_training_file']
    ptf_info = pv.validate_ptf(None, schema_directory, schema_ptfs, True)
    if ptf_info[0] is None or ptf_info[2] is True:
        sys.exit('Please ensure that the schema\'s directory includes the '
                 'Prodigal training file used to create the schema.')

    bsr_val = pv.bsr_type(configs.get('bsr', ''))
    msl_val = pv.minimum_sequence_length_type(configs.get('minimum_locus_length', ''))
    tt_val = pv.translation_table_type(configs.get('translation_table', ''))
    st_val = pv.size_threshold_type(configs.get('size_threshold', ''))
    cv_val = configs.get('chewBBACA_version', '')[0]
    # add window size
    ws_val = pv.validate_ws(configs.get('word_size', None))
    cs_val = pv.validate_cs(configs.get('cluster_sim', None))
    rf_val = pv.validate_rf(configs.get('representative_filter', None))
    if_val = pv.validate_if(configs.get('intraCluster_filter', None))

    # dictionary with schema parameters values to send to the NS
    params = {'bsr': bsr_val, 'prodigal_training_file': ptf_info[1],
              'translation_table': tt_val, 'minimum_locus_length': msl_val,
              'chewBBACA_version': 'chewBBACA {0}'.format(cv_val),
              'size_threshold': st_val, 'word_size': ws_val,
              'cluster_sim': cs_val, 'representative_filter': rf_val,
              'intraCluster_filter': if_val}

    params_values = list(params.values())

    if '' in params_values:
        sys.exit('Found invalid parameters values and exited.')
    else:
        params_lines = ['  {0}: {1}'.format(k, str(v))
                        for k, v in params.items()
                        if k not in ['prodigal_training_file', 'name']]
        params_text = '\n'.join(params_lines)
        print(params_text)
        print('All configurations successfully validated.\n')
        params['name'] = schema_name
        ptf_file = ptf_info[0]
        ptf_hash = ptf_info[1]

    # determine schema status
    upload_type = schema_status(nomenclature_server, headers_get,
                                schema_name, species_id,
                                continue_up)

    if upload_type[0] == 'novel':
        print('New schema name: "{0}" '.format(schema_name))
    else:
        schema_url = '{0}species/{1}/schemas/{2}'.format(nomenclature_server,
                                                         species_id,
                                                         upload_type[1])
        schema_id = schema_url.split('/')[-1]
        print('Schema exists and is incomplete '
              '("{0}", id={1})'.format(schema_name, schema_id))

    # get schema description
    if continue_up is False:
        if description_file is not None and os.path.isfile(description_file) is True:
            # determine file hash
            description_hash = fo.hash_file(description_file, hashlib.blake2b())
            print('Schema description: {0}'.format(description_file))
        else:
            print('Could not get a description from a file. '
                  'Reset to schema name.')
            description_file = 'schema_description.txt'
            with open(description_file, 'w') as sd:
                sd.write(schema_name)
            description_hash = fo.hash_file(description_file, hashlib.blake2b())

        params['SchemaDescription'] = description_hash

    print('\n-- Schema Pre-processing --')

    # hash schema files to get unique identifiers based on content
    hashed_files = {fo.hash_file(file, hashlib.blake2b()): file for file in fasta_paths}
    print('Determining data to upload...')
    absent_loci = fasta_paths
    if upload_type[0] == 'incomplete':

        loci_info, absent_loci, fasta_paths = schema_completedness(nomenclature_server, species_id,
                                                                   upload_type[1], headers_get,
                                                                   hashed_files)

    print('  Loci to create and associate with species '
          'and schema: {0}'.format(len(absent_loci)))
    print('  Loci without the full set of alleles: '
          '{0}\n'.format(len(fasta_paths)))

    # create inputs for QC step
    inputs = [(file,
               file.split('/')[-1].split('.fasta')[0],
               int(params['translation_table']),
               0,
               None) for file in fasta_paths]

    # validate schema data and create files with translated sequences
    print('Translating sequences based on schema configs...')
    qc_results = []
    genes_pools = multiprocessing.Pool(processes=cpu_cores)
    rawr = genes_pools.map_async(quality_control, inputs,
                                 callback=qc_results.extend)
    rawr.wait()

    # get invalid alleles
    invalid_alleles = [r[2] for r in qc_results]
    invalid_alleles = list(itertools.chain.from_iterable(invalid_alleles))
    invalid_identifiers = set([r[0] for r in invalid_alleles])

    print('  Found a total of {0} invalid '
          'alleles.\n'.format(len(invalid_identifiers)))

    # list translated sequences files
    dna_files = [r[0] for r in qc_results]
    prot_files = [r[1] for r in qc_results]

    # determine loci missing annotations
    miss_annotation = [pf for pf in prot_files
                       if pf.split('_prots')[0] + '.fasta' in absent_loci]

    print('Loci missing UniProt annotation: {0}'.format(len(miss_annotation)))

    queries_files = []
    if len(miss_annotation) > 0:
        # create SPARQL queries to query UniProt SPARQL endpoint
        print('Creating SPARQL queries to search UniProt for annotations...')
        with concurrent.futures.ThreadPoolExecutor(max_workers=ct.UNIPROT_SPARQL_THREADS) as executor:
            for res in executor.map(create_uniprot_queries, miss_annotation):
                queries_files.append(res)

        # multithreaded annotation searching
        print('Searching for annotations on UniProt...')
        loci_annotations = {}
        total_found = 0
        total_loci = len(queries_files)
        with concurrent.futures.ThreadPoolExecutor(max_workers=ct.UNIPROT_SPARQL_THREADS) as executor:
            # Start the load operations and mark each future with its URL
            for res in executor.map(get_annotation, queries_files):
                loci_annotations[res[0]] = res[1:]
                total_found += 1
                print('\r', 'Searched annotations for '
                      '{0}/{1} loci'.format(total_found, total_loci), end='')

        # get user and custom annotations
        if os.path.isfile(str(annotations)) is True:
            user_annotations = import_annotations(annotations)
            valid = 0
            for a in loci_annotations:
                if a in user_annotations:
                    loci_annotations[a].extend(user_annotations[a])
                    valid += 1
                else:
                    loci_annotations[a].extend(['N/A', 'N/A'])
            print('\nUser provided valid annotations for {0} '
                  'loci.'.format(valid))
        else:
            loci_annotations = {locus: loci_annotations[locus]+['N/A', 'N/A'] for locus in loci_annotations}
            if annotations is None:
                print('\nUser did not provide file with annotations.')
            else:
                print('\nInvalid annotations file value.')

    # convert parameters to string type because Chewie-NS
    # expects strings
    params = {k: str(v) for k, v in params.items()}
    params['schema_hashes'] = list(hashed_files.keys())

    # start sending data
    print('\n-- Schema Upload --')
    # Build the new schema URL and POST to NS
    if continue_up is False:
        schema_url, schema_id = create_schema(nomenclature_server,
                                              headers_post,
                                              species_id,
                                              params)
        # send file with description
        description_uri = cr.make_url(nomenclature_server, 'species',
                                      species_id, 'schemas', schema_id,
                                      'description')

        desc_res = cr.upload_file(description_file, description_hash,
                                  description_uri, headers_post_bytes,
                                  False)

        print('Created schema with name {0} '
              '(id={1}).\n'.format(schema_name, schema_id))
    else:
        print('Continuing upload of schema with '
              'name {0} (id={1})\n'.format(schema_name, schema_id))

    # start creating new loci and adding/linking alleles
    print('Loci data:')

    if upload_type[0] == 'incomplete':
        if len(absent_loci) == 0:
            response_data = {k: [v[0], True, True, True, v[3]]
                             for k, v in loci_info.items()}
            print('All loci were inserted in a previous process.\n')
        else:
            print('  Collecting loci data...')
            loci_file = create_loci_file(dna_files, loci_annotations,
                                         schema_directory, species_id,
                                         schema_id, loci_prefix,
                                         absent_loci)
            print('  Sending data to chewie-NS...')
            absent_data = upload_loci_data(loci_file, nomenclature_server,
                                           species_id, schema_id,
                                           headers_post_bytes, headers_get,
                                           hashed_files)
            response_data = {k: [v[0], True, True, True, v[3]]
                             for k, v in loci_info.items() if k not in absent_data}
            response_data = {**absent_data, **response_data}
            print('\n  The NS completed the insertion of {0} '
                  'loci.\n'.format(len(absent_loci)))
    else:
        print('  Collecting loci data...')
        loci_file = create_loci_file(dna_files, loci_annotations,
                                     schema_directory, species_id,
                                     schema_id, loci_prefix)
        print('  Sending data to chewie-NS...')
        response_data = upload_loci_data(loci_file, nomenclature_server,
                                         species_id, schema_id,
                                         headers_post_bytes, headers_get,
                                         hashed_files)

        print('\n  The NS completed the insertion of {0} '
              'loci.\n'.format(len(response_data)))

    # create files with info for posting alleles
    print('Alleles data:')
    print('  Collecting alleles data...')
    (alleles_files, loci_ids, loci_hashes,
     loci_names) = create_alleles_files(dna_files, response_data,
                                        invalid_identifiers, species_name,
                                        nomenclature_server, species_id,
                                        schema_id, user_id)
    # determine length of all alleles per locus
    length_files = create_lengths_files(dna_files, schema_directory)

    # zip all files to reduce upload size
    print('  Compressing files with alleles data...')
    zipped_files = ['{0}.zip'.format(file) for file in alleles_files]
    list(map(fo.file_zipper, alleles_files, zipped_files))
    alleles_data = list(zip(zipped_files, loci_ids, loci_hashes, loci_names))

    print('  Sending alleles data to chewie-NS...')
    # send POST with file contents and process each file in the NS
    failed = upload_alleles_data(alleles_data, length_files,
                                 nomenclature_server, headers_post,
                                 headers_post_bytes, species_id,
                                 schema_id)

    if len(failed) > 0:
        sys.exit('Could not upload data for alleles of following loci:\n'
                 '{0}\n Please retry with the "--continue_up" option and '
                 'contact the NS Admin if the problem '
                 'persists.'.format(','.join(failed)))
    else:
        # send training file to NS
        print('\n\nUploading Prodigal training file...')
        ptf_url = cr.make_url(nomenclature_server, 'species', species_id,
                              'schemas', schema_id, 'ptf')
        ptf_res = cr.upload_file(ptf_file, ptf_hash,
                                 ptf_url, headers_post_bytes,
                                 False)
        print(list(ptf_res.json().values())[0])
        print('\nchewie-NS has received the data and will insert '
              'the alleles into the database.')
        print('Schema will be available for download as soon as '
              'the process has completed.')
        print('Schema information will also be available on the '
              'chewie-NS website.\n')

    # delete all intermediate files
    print('Removing intermediate files...')
    fo.remove_files(prot_files)
    fo.remove_files(queries_files)
    fo.remove_files(length_files)
    fo.remove_files(alleles_files)
    fo.remove_files(zipped_files)

    if len(absent_loci) > 0:
        os.remove(loci_file)
        os.remove('{0}.zip'.format(loci_file))


def parse_arguments():

    parser = argparse.ArgumentParser(description=__doc__,
                                     formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument('-i', type=str, required=True,
                        dest='schema_directory',
                        help='Path to the directory of the schema to upload.')

    parser.add_argument('-sp', type=str, required=True,
                        dest='species_id',
                        help='The integer identifier or name of the species '
                             'that the schema will be associated to in '
                             'the NS.')

    parser.add_argument('-sn', type=str, required=True,
                        dest='schema_name',
                        help='A brief and meaningful name that '
                             'should help understand the type and content '
                             'of the schema.')

    parser.add_argument('-lp', type=str, required=True,
                        dest='loci_prefix',
                        help='Prefix included in the name of each locus of '
                             'the schema.')

    parser.add_argument('--df', type=str, required=False,
                        dest='description_file', default=None,
                        help='Path to a text file with a description '
                             'about the schema. Markdown syntax is supported '
                             'in order to offer greater customizability of '
                             'the rendered description in the Frontend. '
                             'Will default to the schema\'s name if the user '
                             'does not provide a valid path for a file.')

    parser.add_argument('--a', type=str, required=False,
                        dest='annotations', default=None,
                        help='Path to a TSV file with loci annotations. '
                             'The first column has loci identifiers '
                             '(w/o .fasta extension), the second has user '
                             'annotations and the third has custom '
                             'annotations.')

    parser.add_argument('--cpu', type=int, required=False,
                        dest='cpu_cores', default=1,
                        help='Number of CPU cores that will '
                             'be used in the Schema Pre-processing step.')

    parser.add_argument('--ns', type=pv.validate_ns_url, required=False,
                        default='main',
                        dest='nomenclature_server',
                        help='The base URL for the Nomenclature Server. '
                             'The default value, "main", will establish a '
                             'connection to "https://chewbbaca.online/", '
                             '"tutorial" to "https://tutorial.chewbbaca.online/" '
                             'and "local" to "http://127.0.0.1:5000/NS/api/" (localhost). '
                             'Users may also provide the IP address to other '
                             'Chewie-NS instances.')

    parser.add_argument('--continue_up', required=False, action='store_true',
                        dest='continue_up',
                        help='If the process should check if the schema '
                             'upload was interrupted and try to resume it.')

    args = parser.parse_args()

    return args


if __name__ == "__main__":

    args = parse_arguments()
    main(**vars(args))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Purpose
-------

This module removes a set of loci from a TSV file with
results from the AlleleCall process.

Code documentation
------------------
"""


import csv
import argparse

import pandas as pd

try:
    from utils import file_operations as fo
except ModuleNotFoundError:
    from CHEWBBACA.utils import file_operations as fo


def main(input_file, genes_list, output_file, inverse):

    # read genes list
    with open(genes_list, 'r') as infile:
        genes_list = list(csv.reader(infile, delimiter='\t'))
        genes_list = [g[0] for g in genes_list[1:]]

    # get list of loci in allele call results
    loci = fo.read_lines(input_file, strip=True, num_lines=1)
    loci = loci[0].split('\t')[1:]
    print('Total loci: {0}'.format(len(loci)))

    if inverse is True:
        columns_to_keep = [g for g in loci if g in genes_list]
    else:
        columns_to_keep = [g for g in loci if g not in genes_list]

    columns_to_remove = (len(loci)) - len(columns_to_keep)
    print('Loci to remove: {0}'.format(columns_to_remove))

    # include first column with sample ids
    columns_to_keep = ['FILE'] + columns_to_keep
    df = pd.read_csv(input_file, usecols=columns_to_keep,
                     sep='\t', dtype=str)

    # save dataframe to file
    df.to_csv(output_file, header=True, sep='\t', index=False)


def parse_arguments():

    parser = argparse.ArgumentParser(description=__doc__,
                                     formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument('-i', '--input-file', type=str,
                        required=True, dest='input_file',
                        help='TSV file that contains a matrix with '
                             'allelic profiles determined by the '
                             'AlleleCall process.')

    parser.add_argument('-gl', '--genes-list', type=str,
                        required=True, dest='genes_list',
                        help='File with the list of genes to '
                             'remove, one identifier per line.')

    parser.add_argument('-o', '--output-file', type=str,
                        required=True, dest='output_file',
                        help='Path to the output file.')

    parser.add_argument('--inverse', action='store_true',
                        required=False, dest='inverse',
                        help='List of genes that is provided '
                             'is the list of genes to keep and '
                             'all other genes should be removed.')

    args = parser.parse_args()

    return args


if __name__ == "__main__":

    args = parse_arguments()
    main(**vars(args))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Purpose
-------

This module contains functions to perform requests to
Chewie-NS (https://github.com/B-UMMI/Chewie-NS).

Code documentation
------------------
"""


import sys
import json
import requests
from getpass import getpass
from urllib.parse import urlparse, urlencode, urlsplit, parse_qs

try:
    from utils import constants as ct
except ModuleNotFoundError:
    from CHEWBBACA.utils import constants as ct


def simple_get_request(url, headers, endpoint_list=None,
                       parameters=None, verify=False, timeout=30):
    """Request data through a GET method sent to an endpoint.

    Construct an URI for an endpoint and request data through
    a GET method sent to that endpoint.

    Parameters
    ----------
    url : str
        The base URL for a Chewie-NS instance.
    headers : dict
        HTTP headers for GET requests.
    endpoint_list : list
        List with elements that will be concatenated to the
        base URI to create the URI for the API endpoint.
    parameters : dict or list or bytes
        Dictionary, list of tuples or bytes to send in the
        query string for the Request.
    verify : bool
        If the SSL certificates should be verified in HTTPS
        requests (False for no verification, True otherwise).
    timeout : int
        Maximum time in seconds to wait for response.

    Returns
    -------
    url : str
        Complete URI for the API endpoint.
    res : requests.models.Response
        The response returned through the GET method.
    """
    # unpack list of sequential endpoints and pass to create URI
    if endpoint_list is not None:
        url = make_url(url, *endpoint_list)

    res = requests.get(url, headers=headers, params=parameters,
                       timeout=timeout, verify=verify)

    # return queried endpoint URI and response
    return [url, res]


def simple_post_request(url, headers, endpoint_list=None,
                        data=None, files=None, verify=False):
    """Send data through a POST method.

    Construct an URI for an endpoint and send data through
    a POST method.

    Parameters
    ----------
    url : str
        The base URL for a Chewie-NS instance.
    headers : dict
        HTTP headers for POST requests.
    endpoint_list : list
        List with elements that will be concatenated to the
        base URI to create the URI for the API endpoint.
    data
        Data to send. Must be an object that can be JSON
        serialized.
    files : dict
        Dictionary with file data to send (see:
        https://requests.readthedocs.io/en/master/user
        /quickstart/#post-a-multipart-encoded-file).
    verify : bool
        If the SSL certificates should be verified in HTTPS
        requests (False for no verification, True otherwise).

    Returns
    -------
    url : str
        Complete URI for the API endpoint.
    res : requests.models.Response
        The response returned through the POST method.
    """
    # unpack list of sequential endpoints and pass to create URI
    if endpoint_list is not None:
        url = make_url(url, *endpoint_list)

    # if the content is defined as 'application/json'
    # the data has to be passed after being processed with
    # json.dumps()
    res = requests.post(url, headers=headers, data=data,
                        files=files, verify=verify)

    return [url, res]


def check_connection(url, headers=ct.HEADERS_GET_JSON):
    """Verify connection to a Chewie-NS instance.

    Perform a GET request to the stats/summary endpoint of a
    Chewie-NS instance to determine if it is possible to
    establish a connection.

    Parameters
    ----------
    url : str
       The base URL for a Chewie-NS instance.
    headers : dict
        HTTP headers for GET requests.

    Returns
    -------
    connection : bool
        True if the GET request to the stats/summary
        endpoint was successful (HTTP response status
        code 200 or 201), False otherwise.
    """
    try:
        url, res = simple_get_request(url, headers, ['stats', 'summary'])
        server_status = res.status_code
        if server_status in [200, 201]:
            connection = True
        else:
            connection = False
    except Exception:
        connection = False

    return connection


def user_info(base_url, headers_get):
    """Get the information about a registered user in Chewie-NS.

    Get the user identifier, role and permission level for a
    user registered in a Chewie-NS instance. The permission
    level is used to determine if the user can upload schemas
    to Chewie-NS.

    Parameters
    ----------
    base_url : str
        The base URL for a Chewie-NS instance.
    headers_get : dict
        HTTP headers for GET requests.

    Returns
    -------
    user_id : str
        User identifier in the Chewie-NS instance.
    user_role : str
        User role in the Chewie-NS instance.
    permission : bool
        True if the user has Admin or Contributor privileges,
        False otherwise.
    """
    # verify user role to check permission
    user_url, user_info = simple_get_request(base_url, headers_get,
                                             ['user', 'current_user'])
    user_info = user_info.json()

    user_id = str(user_info['id'])
    user_role = user_info['roles'].split(' ')[-1][:-1]
    permission = any(role in user_role for role in ['Admin', 'Contributor'])

    return [user_id, user_role, permission]


def species_ids(species_id, base_url, headers_get):
    """Get a species identifier or name in Chewie-NS.

    Parameters
    ----------
    species_id : str
        The identifier of the species in Chewie-NS or the
        name of the species.
    base_url : str
        The base URL for a Chewie-NS instance.
    headers_get : dict
        HTTP headers for GET requests.

    Returns
    -------
    A list with the species identifier and name if the species
    exists in Chewie-NS or a 404 HTTP status code if the species
    was not found.
    """
    try:
        # check if species unique integer identifier was provided
        int(species_id)
        species_url, species_info = simple_get_request(base_url, headers_get,
                                                       ['species', species_id])
        if species_info.status_code in [200, 201]:
            species_name = species_info.json()[0]['name']['value']
            return [species_id, species_name]
        else:
            return 404
    except ValueError:
        # check if the scientific name for species was provided
        species_name = species_id
        # get the list of species added to Chewie-NS
        ns_species = species_list(base_url, headers_get, ['species', 'list'])
        # try to get the species unique integer identifier
        species_id = ns_species.get(species_name, 'not_found')
        if species_id != 'not_found':
            return [species_id, species_name]
        else:
            return 404


def species_list(base_url, headers_get, endpoint_list):
    """Get the list of species in a Chewie-NS instance.

    Parameters
    ----------
    base_url : str
        The base URL for a Chewie Nomenclature Server
        instance.
    headers_get : dict
        HTTP headers for GET requests.
    endpoint_list : list
        List with elements that will be concatenated
        to the base URI to create the URI for the API endpoint.

    Returns
    -------
    species_lst : dict
        Dictionary with species names as keys and
        species identifiers as values.
    """
    species_url, res = simple_get_request(base_url, headers_get, endpoint_list)
    res = res.json()
    species_lst = {}
    for sp in res:
        # get species scientific name
        species = sp['name']['value']
        # get species unique integer identifier from URl
        species_url = sp['species']['value']
        species_id = species_url.split('/')[-1]
        species_lst[species] = species_id

    return species_lst


def is_url(url):
    """Check if a URL is valid.

    Parameters
    ----------
    url : str
        The URL to be validated.

    Returns
    -------
    True if the URL is valid, False otherwise.
    """
    try:
        result = urlparse(url)
        return all([result.scheme, result.netloc, result.path])
    except Exception:
        return False


def make_url(base_url, *res, **params):
    """Create a URL.

    Parameters
    ----------
    base_url : str
        The base URL.
    res : list
        List with elements that will be concatenated to the
        base URI to create the URI for the API endpoint.
    params : str
        Addtional parameters (WIP).

    Returns
    -------
    url : str
        URL constructed by adding endpoints and additional
        parameters. If the provided base URL is invalid
        returns 'An invalid URL was provided.'
    """
    url = base_url
    # Check if the url is valid
    if is_url(url):

        if url[-1] == '/':
            url = url[:-1]

        # Add the endpoints
        for r in res:
            url = '{0}/{1}'.format(url, r)

        # Add params if they are provided
        if params:
            url = '{0}?{1}'.format(url, urlencode(params))

        return url
    else:
        return 'An invalid URL was provided.'


def get_sequence_from_url(url):
    """Extract sequence from URL query component.

    Parameters
    ----------
    url : str
        URL that contains the sequence.

    Returns
    -------
    seq : str
        Sequence extracted from the URL.
    """
    seq = parse_qs(urlsplit(url).query)['sequence'][0]

    return seq


def login_user_to_NS(server_url, email, password):
    """Login user to a Chewie-NS instance.

    Parameters
    ----------
    server_url : str
        The base URL for a Chewie-NS instance.
    email : str
        Email linked to user's account in Chewie-NS.
    password : str
        Password for the user's account in Chewie-NS.

    Returns
    -------
    token : str
        Authorization token to perform requests to Chewie-NS
        instance if login is successful. Otherwise, returns
        False.
    """
    auth_params = {}
    auth_params['email'] = email
    auth_params['password'] = password

    auth_url, auth_r = simple_post_request(server_url,
                                           ct.HEADERS_POST_JSON,
                                           ['auth', 'login'],
                                           data=json.dumps(auth_params))

    auth_result = auth_r.json()
    if auth_result['status'] == 'success':
        token = auth_result['access_token']
    else:
        token = False

    return token


def capture_login_credentials(base_url):
    """Capture login credentials to login user into Chewie-NS.

    Parameters
    ----------
    base_url : str
        The base URL for a Chewie-NS instance.

    Returns
    -------
    token : str
        Authorization token to perform requests to Chewie-NS.
    """
    print('\nPlease provide login credentials:')
    user = input('USERNAME: ')
    password = getpass('PASSWORD: ')
    print()
    # get token
    token = login_user_to_NS(base_url, user, password)
    # if login was not successful, stop the program
    if token is False:
        sys.exit('Invalid credentials.')

    return token


def get_species_schemas(schema_id, species_id, base_url, headers_get):
    """Determine if there is a schema with specified identifier.

    Retrieves the list of schemas for a species in a Chewie-NS
    instance and determines if it has a schema with the
    specified identifier.

    Parameters
    ----------
    schema_id : str
        The identifier of the schema in Chewie-NS.
    species_id : str
        The identifier of the schema's species in Chewie-NS.
    base_url : str
        The base URL for the Chewie-NS instance.
    headers_get : dict
        HTTP headers for GET requests.

    Returns
    -------
    A list with the schema identifier, the schema URI, the
    schema name and a dictionary with all properties for
    the schema in Chewie-NS.

    Raises
    ------
    SystemExit
        - If the schema with the specified identifier does
          not exist.
        - If the process cannot retrieve the list of schemas
          for the species.
    """
    # get the list of schemas for the species
    schemas_url, schema_get = simple_get_request(base_url, headers_get,
                                                 ['species', species_id,
                                                  'schemas'])
    schema_get_status = schema_get.status_code
    if schema_get_status in [200, 201]:
        species_schemas = schema_get.json()

        # extract schemas identifiers, URIs and names from response
        schemas_info = []
        for s in species_schemas:
            schema_uri = s['schemas']['value']
            scid = schema_uri.split('/')[-1]
            schema_name = s['name']['value']
            schemas_info.append([scid, schema_uri, schema_name])

        # select schema with specified identifier
        schema = [s for s in schemas_info if schema_id in s]
        if len(schema) > 0:
            # get schema parameters
            schema = schema[0]
            url, schema_params = simple_get_request(schema[1], headers_get)
            schema_params = schema_params.json()[0]
            schema.append(schema_params)
            return schema
        else:
            sys.exit('Could not find a schema with provided identifier.')
    else:
        sys.exit('Could not retrieve schemas for current species.')


def upload_file(file, filename, url, headers, verify_ssl):
    """Upload a file to a Chewie-NS instance.

    Parameters
    ----------
    file : str
        Path to the file to upload.
    filename : str
        Name used to save the file in Chewie-NS.
    url : str
        Endpoint URL that receives the POST request.
    headers : dict
        HTTP POST request headers.
    verify_sll : bool
        If the SSL certificates should be verified in
        HTTPS requests (False for no verification, True
        otherwise).

    Returns
    -------
    response : requests.models.Response
        Response object from the 'requests' module.
    """
    file_handle = open(file, 'rb')
    files = {'file': (filename, file_handle)}
    up_url, response = simple_post_request(url, headers,
                                           files=files,
                                           verify=verify_ssl)

    file_handle.close()

    return response



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Purpose
-------

This module determines the set of genes in the core genome based on
a matrix with allelic profiles and a threshold that defines the
proportion of genomes a gene must be present in to be included in
the core genome.

Expected input
--------------

The process expects the following variables whether through command line
execution or invocation of the :py:func:`main` function:

- ``-i``, ``input_file`` : Path to input file containing a matrix with
     'allelic profiles.

    - e.g.: ``/home/user/chewie/results/matrix``

- ``-o``, ``output_directory`` : Path to the directory where the process
  will store output files.

    - e.g.: ``/home/user/chewie/results/output_directory``

- ``--t``, ``threshold`` : Genes that constitute the core genome must be
  in a proportion of genomes that is at least equal to this value.

    - e.g.: ``0.95``

- ``--s``, ``step`` : Number of genomes added to the cgMLST computation
  at each step.

    - e.g.: ``5``

- ``--r``, ``genes2remove`` : Path to file with a list of genes/columns to
  remove from the matrix (one gene identifier per line).

    - e.g.: ``home/user/results/genes.txt``

- ``--g``, ``genomes2remove`` : Path to file with a list of genomes/rows
  to remove from the matrix (one genome identifier per line).

Code documentation
------------------
"""


import os
import numpy as np
import pandas as pd
from plotly.offline import plot
import plotly.graph_objects as go

try:
    from utils import (file_operations as fo,
                       iterables_manipulation as im)
except ModuleNotFoundError:
    from CHEWBBACA.utils import (file_operations as fo,
                                 iterables_manipulation as im)


def binarize_matrix(column):
    """Convert a Pandas dataframe column values into numeric values.

    Parameters
    ----------
    column : pandas.core.series.Series
        Pandas dataframe column.

    Returns
    -------
    Numpy array corresponding to the input column
    with numeric values equal to 1 for cells that
    had valid allele identifiers and equal to 0
    for cells that had missing data.
    """
    coln = pd.to_numeric(column)

    return np.int64(coln > 0)


def remove_genomes(matrix, genomesToRemove):
    """Remove rows from an allele calling matrix.

    Remove rows from a Pandas dataframe if the
    index identifier matches the identifier of
    a genome to remove.

    Parameters
    ----------
    matrix : pandas.core.frame.DataFrame
        Pandas dataframe with allelic profiles.
        Each row has the allelic profile of a genome
        and each column has the allele identifiers
        determined for a locus.
    genomesToRemove : list
        List of genomes to remove.

    Returns
    -------
    pruned_matrix : pandas.core.frame.DataFrame
        Input dataframe without the rows whose
        index matched an identifier of a genome
        to remove.
    """
    # determine row indexes that match any genome to remove
    to_remove_bool = matrix.index.isin(genomesToRemove)
    # create new matrix without rows that matched any genome to remove
    pruned_matrix = matrix.loc[~ to_remove_bool]

    return pruned_matrix


def remove_genes(matrix, genesToRemove):
    """Remove columns from an allele calling matrix.

    Parameters
    ----------
    matrix : pandas.core.frame.DataFrame
        Pandas dataframe with allelic profiles.
        Each row has the allelic profile of a genome
        and each column has the allele identifiers
        determined for a locus.
    genesToRemove : list
        List of loci to remove.
    """
    matrix = matrix[matrix.columns[~matrix.columns.isin(genesToRemove)]]

    return matrix


def above_threshold(column, column_length, threshold):
    """Determine if gene presence is equal or above a threshold.

    Parameters
    ----------
    column : pandas.core.series.Series
        Pandas dataframe column with presence (1) and absence (0)
        values for a locus in a set of genomes.
    column_length : int
        Number of genomes in the dataset.
    threshold : float
        Core genome determination threshold.

    Returns
    -------
    bool
        True if gene is equal or above threshold, False otherwise.
    """
    return (np.sum(column) / column_length) >= threshold


def compute_cgMLST(matrix, sorted_genomes, threshold, step):
    """Prune allele call results based on presence/absence threshold.

    Removes columns from an allele calling matrix based on
    threshold for loci presence/absence.

    Parameters
    ----------
    matrix : pandas.core.frame.DataFrame
        Pandas dataframe with allelic profiles.
        Each row has the allelic profile of a genome
        and each column has the allele identifiers
        determined for a single gene.
    sorted_genomes : list
        List of genome identifiers sorted in order of
        decresing number of loci.
    threshold : float
        Core genome determination threshold.
    step : int
        Number of genomes added to the cgMLST computation at
        each step.

    Returns
    -------
    pruned_df : pandas.core.frame.DataFrame
        Dataframe with the cgMLST profiles for the last step value
        (the last step includes all genomes in the input matrix).
    cgMLST_size : dict
        Dictionary with the number of genomes used to compute the
        cgMLST as keys and the size of the core-genome as values.
    """
    # determine genes at or above threshold
    cgMLST_size = {}
    for i in im.inclusive_range(1, len(sorted_genomes), step):
        # get subdataframe for current genomes
        current_df = matrix.loc[sorted_genomes[:i]]
        pa_rows, _ = current_df.shape
        is_above_threshold = current_df.apply(above_threshold,
                                              args=(pa_rows, threshold,))
        above = current_df.columns[is_above_threshold]
        cgMLST_size[pa_rows] = len(above)
        print('\r', 'Computed for...{0} genomes.'.format(i), end='')

    # return list of genes in cgMLST and cgMLST count per genome threshold
    return [above, cgMLST_size]


def presAbs(matrix, output_directory):
    """Create a presence/absence matrix.

    Parameters
    ----------
    matrix : pandas.core.frame.DataFrame
        Pandas dataframe with allelic profiles.
        Each row has the allelic profile of a genome
        and each column has the allele identifiers
        determined for a single gene.
    output_directory : str
        Path to the directory where the TSV file with
        the presence absence matrix will be stored.

    Returns
    -------
    presence_absence : pandas.core.frame.DataFrame
        Pandas dataframe with numeric values equal to
        1 for the cells that had valid allele identifiers
        and equal to 0 for missing data.
    pa_path : str
        Path to the output TSV file that contains the
        presence-absence matrix.
    """
    presence_absence = matrix.apply(binarize_matrix)

    pa_path = fo.join_paths(output_directory, ['Presence_Absence.tsv'])
    presence_absence.to_csv(pa_path, sep='\t')

    return [presence_absence, pa_path]


def missing_data_table(presence_absence):
    """Determine missing data per genome.

    Parameters
    ----------
    presence_absence : pandas.core.frame.DataFrame
        Pandas dataframe with numeric values equal to
        1 for the cells that have valid allele identifiers
        and equal to 0 for missing data.

    Returns
    -------
    missing_data_df : pandas.core.frame.DataFrame
        Dataframe with number of missing genes and
        percentage of missing genes per genome.
    """
    _, n_genes = presence_absence.shape
    genes_present = presence_absence.apply(np.count_nonzero, axis=1)

    missing_data = {'FILE': presence_absence.index,
                    'missing': n_genes - genes_present,
                    'percentage': 1 - (genes_present / n_genes)}

    missing_data_df = pd.DataFrame(missing_data,
                                   columns=['FILE',
                                            'missing',
                                            'percentage'])

    return missing_data_df


def main(input_file, output_directory, threshold, step,
         genes2remove, genomes2remove):
    """Determine the cgMLST based on allele calling results.

    Parameters
    ----------
    input_file : str
        Path a TSV file with allelic profiles for a set
        of genomes.
    output_directory : str
        Path to the directory where the process will
        store output files.
    threshold : list
        Core genome determination thresholds.
    step : int
        Number of genomes added to the cgMLST computation at
        each step.
    genes2remove : str
        Path to TXT file with the list of genomes to remove.
    genomes2remove : str
        Path to TXT file with the list of loci to remove.

    Returns
    -------
    List with the paths to three files:

    - Path a TSV file with the cgMLST matrix.
    - Path to a TXT file with the list of genes that
      constitute the core genome.
    - Path to a TSV file with the information about
      missing data per genome.
    """
    fo.create_directory(output_directory)

    # import matrix with allelic profiles
    matrix = pd.read_csv(input_file, header=0, index_col=0,
                         sep='\t', low_memory=False)

    total_genomes, total_loci = matrix.shape
    print('Input matrix has {0} profiles for {1} '
          'loci.'.format(total_genomes, total_loci))

    # read list of genes and list of genomes to exclude
    genesToRemove = []
    if genes2remove:
        genesToRemove = fo.read_lines(genes2remove)
    print('{0} loci to exclude.'.format(len(genesToRemove)))
    genomesToRemove = []
    if genomes2remove:
        genomesToRemove = fo.read_lines(genomes2remove)
    print('{0} genomes to exclude.'.format(len(genomesToRemove)))

    cgMLST_thresholds = threshold

    # remove genomes
    if len(genomesToRemove) > 0:
        matrix = remove_genomes(matrix, genomesToRemove)
        total_genomes -= len(genomesToRemove)

    # remove genes
    if len(genesToRemove) > 0:
        matrix = remove_genes(matrix, genesToRemove)
        total_loci -= len(genesToRemove)

    # mask missing data
    print('Masking matrix with {0} profiles for {1} '
          'loci...'.format(total_genomes, total_loci))
    masked_matrix = matrix.apply(im.replace_chars)
    print('Masked {0} profiles.'.format(total_genomes))

    # build presence/absence matrix
    print('Building presence-absence matrix...')
    pa_matrix, pa_outfile = presAbs(masked_matrix, output_directory)
    print('Presence-absence matrix saved to {0}'.format(pa_outfile))

    # count number of missing data per genome
    print('Determining missing data per genome...')
    missing_data_df = missing_data_table(pa_matrix)
    missing_stats = missing_data_df['missing'].describe()

    # sort genomes based on order of decreasing missing data
    missing_data_df = missing_data_df.sort_values('missing', ascending=True)
    sorted_genomes = missing_data_df.index.tolist()

    # write table with missing data stats
    mdata_path = os.path.join(output_directory, 'mdata_stats.tsv')
    missing_data_df.to_csv(mdata_path, sep='\t', index=False)
    print('Missing data table saved to {0}'.format(mdata_path))

    # compute cgMLST
    line_traces = []
    for i, t in enumerate(cgMLST_thresholds):
        print('Determining cgMLST for loci presence threshold of {0}...'.format(t))
        cgMLST_genes, cgMLST_counts = compute_cgMLST(pa_matrix,
                                                     sorted_genomes,
                                                     t,
                                                     step)

        retained = len(cgMLST_genes)
        print('\ncgMLST for loci presence threshold of {0} composed '
              'of {1}/{2} genes.'.format(t, retained, total_loci))

        # write cgMLST matrix
        # get subset from masked matrix
        cgMLST_matrix = masked_matrix[cgMLST_genes]
        cgmlst_path = os.path.join(output_directory, 'cgMLST{0}.tsv'.format(int(t*100)))
        cgMLST_matrix.to_csv(cgmlst_path, sep='\t')
        print('cgMLST profiles saved to {0}'.format(cgmlst_path))

        # write genes in cgMLST to file
        loci_path = os.path.join(output_directory, 'cgMLSTschema{0}.txt'.format(int(t*100)))
        fo.write_lines(list(cgMLST_genes), loci_path)
        print('List of loci in the cgMLST saved to {0}'.format(loci_path))

        # create line trace
        cgMLST_trace = go.Scattergl(x=list(cgMLST_counts.keys()),
                                    y=list(cgMLST_counts.values()),
                                    mode='lines',
                                    name='cgMLST{0}'.format(int(t*100)),
                                    hovertemplate=('%{y}'))
        line_traces.append(cgMLST_trace)

    # create trace with present loci
    present = [total_loci-i for i in missing_data_df['missing']]
    genomes_index = list(range(1, len(present)+1))
    miss_trace = go.Scattergl(x=genomes_index,
                              y=present,
                              mode='lines',
                              name='Present in genome',
                              line=dict(color='#000000'),
                              hovertemplate=('%{y}<br>'
                                             'Genome: %{text}<br>'),
                              text=sorted_genomes)
    line_traces.append(miss_trace)

    fig = go.Figure(data=line_traces)
    fig.update_layout(title={'text': 'Number of loci in cgMLST',
                             'font_size': 30},
                      xaxis_title='Number of genomes',
                      yaxis_title='Number of loci',
                      template='simple_white',
                      hovermode='x')
    fig.update_xaxes(range=[0, len(sorted_genomes)],
                     tickfont=dict(size=18),
                     titlefont=dict(size=20),
                     showgrid=True)
    fig.update_yaxes(tickfont=dict(size=18),
                     titlefont=dict(size=20),
                     showgrid=True)
    output_html = os.path.join(output_directory, 'cgMLST.html')
    plot(fig, filename=output_html, auto_open=False)
    print('HTML file with cgMLST per loci presence threshold '
          'and per step saved to {0}'.format(output_html))


if __name__ == '__main__':

    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Purpose
-------

This module allows users to synchronize a local schema, previously
downloaded from the Chewie-NS, with the remote schema in the Chewie-NS.

The synchronization process will retrieve new alleles added to the remote
schema after the last synchronization date and update the local schema,
ensuring that local and remote alleles have the same integer identifiers.
The process also allows users to submit novel alleles and update the remote
schema. Novel alleles that are not submitted keep a '*' in the identifier.

Code documentation
------------------
"""


import os
import sys
import json
import time
import shutil
import requests
from collections import defaultdict
from urllib3.exceptions import InsecureRequestWarning

from Bio import SeqIO

try:
    from PrepExternalSchema import PrepExternalSchema
    from utils import (constants as ct,
                       file_operations as fo,
                       chewiens_requests as cr,
                       # profiles_sqlitedb as ps,
                       parameters_validation as pv,
                       iterables_manipulation as im)
except ModuleNotFoundError:
    from CHEWBBACA.PrepExternalSchema import PrepExternalSchema
    from CHEWBBACA.utils import (constants as ct,
                                 file_operations as fo,
                                 chewiens_requests as cr,
                                 # profiles_sqlitedb as ps,
                                 parameters_validation as pv,
                                 iterables_manipulation as im)


# Suppress only the single warning from urllib3 needed.
requests.packages.urllib3.disable_warnings(category=InsecureRequestWarning)


def create_lengths_files(upload, temp_dir):
    """Create files with sequence length values to send to Cheiwe-NS.

    Parameters
    ----------
    upload : dict
        Dictionary with loci identifiers as keys and lists
        as values. Each list contains the following elements:

        - A dictionary with sequences hashes as keys and a list
          with the sequence identifier, DNA sequence and sequence
          length value.
        - The locus URI.
    temp_dir : str
        Path to the directory where the output files with
        the length values of the sequences will be created.

    Returns
    -------
    length_files : list of str
        List with paths to the pickled files that contain
        the sequence length values. Each pickled file
        contains a dictionary with hashes (sequence hashes
        computed with sha256) as keys and sequence lengths
        as values.
    """
    length_files = []
    for locus, v in upload.items():
        records = v[0]
        locus_id = locus.rstrip('.fasta')
        lengths = {locus: {h: info[2] for h, info in records.items()}}

        lengths_file = fo.join_paths(temp_dir,
                                     ['{0}_lengths'.format(locus_id)])
        fo.pickle_dumper(lengths, lengths_file)
        length_files.append(lengths_file)

    return length_files


def create_alleles_files(upload, base_url, user_id, species_name,
                         species_id, schema_id, temp_dir):
    """Create files with the essential data to insert alleles in Chewie-NS.

    Parameters
    ----------
    upload : dict
        Dictionary with loci identifiers as keys and lists
        as values. Each list contains the following elements:

        - A dictionary with sequences hashes as keys and a list
          with the sequence identifier, DNA sequence and sequence
          length value.
        - The locus URI.
    base_url : str
        Base URL of the Chewie-NS.
    user_id : int
        Current user identifier.
    species_name : str
        Name of the species.
    species_id : int
        The identifier of the schema's species in the NS.
    schema_id : int
        The identifier of the schema in the NS.
    temp_dir : str
        Path to the directory where the output files  will
        be created.

    Returns
    -------
    list of list
        List with the following elements:

        - List with paths to files that contain the data that
          is necessary to upload and insert new alleles.
        - List with the loci identifiers in the NS.
        - List with loci basenames.
    """
    loci_ids = []
    loci_names = []
    alleles_files = []
    user_uri = '{0}users/{1}'.format(base_url, user_id)
    for locus, recs in upload.items():
        locus_uri = recs[1]
        alleles_sequences = [v[1] for k, v in recs[0].items()]

        post_inputs = [locus_uri, species_name,
                       user_uri, tuple(alleles_sequences)]

        locus_id = locus_uri.split('/')[-1]
        loci_ids.append(locus_id)

        locus_name = locus.rstrip('.fasta')
        loci_names.append(locus_name)

        alleles_file = fo.join_paths(temp_dir,
                                     ['{0}_{1}_{2}'.format(species_id,
                                                           schema_id,
                                                           locus_id)])
        alleles_files.append(alleles_file)

        fo.pickle_dumper(post_inputs, alleles_file)

    return [alleles_files, loci_ids, loci_names]


def upload_alleles_data(alleles_data, length_files, base_url, headers_post,
                        headers_post_bytes, species_id, schema_id):
    """Upload files with the data to insert alleles in Chewie-NS.

    Parameters
    ----------
    alleles_data : list
        List with tuples, one per locus, that contain the path
        to the ZIP archive with the data to insert alleles,
        the identifier of the locus, the locus file hash and
        the basename of the locus file.
    length_files : list
        List with paths to the pickled files that contain a
        dictionary with sequences hashes as keys and sequence
        length as values.
    base_url : str
        Base URL of the Nomenclature server.
    headers_post : dict
        HTTP headers for POST requests that accept JSON
        formatted data.
    headers_post_bytes : dict
        HTTP headers for POST requests that support file
        upload.
    species_id : int
        The identifier of the schema's species in the NS.
    schema_id : int
        The identifier of the schema in the NS.

    Returns
    -------
    failed : list of str
        List with the identifiers of the loci whose alleles
        data could not be fully uploaded.
    zip_res : dict
        A dictionary with the response returned by the last
        POST method. It has loci identifiers as keys and
        lists with two dictionaries as values (the dictionaries
        have sequences hashes as keys and sequence identifiers in
        the Chewie-NS as values. The first dictionary has the hashes
        of the sequences that were sent to the Chewie-NS but that were
        already present in the loci and the identifiers of those repeated
        alleles that were sent to the Chewie-NS. The second dictionary
        has the same structure but for the sequences that were accepted and
        inserted into each locus).
    """
    uploaded = 0
    failed = []
    for i, a in enumerate(alleles_data):
        locus_id = a[1]
        # get length of alleles from current locus
        current_len = length_files[i]
        data = fo.pickle_loader(current_len)
        data = {locus_id: data[next(iter(data))]}
        data = {'content': data}

        # send data to the NS
        send_url = cr.make_url(base_url, 'species', species_id,
                               'schemas', schema_id, 'loci',
                               locus_id, 'lengths')

        lengths_res = cr.simple_post_request(send_url, headers_post,
                                             data=json.dumps(data))[1]
        length_status = lengths_res.status_code

        # get path to ZIP archive with data to insert alleles
        current_zip = a[0]

        # send data to insert alleles in the NS
        zip_url = cr.make_url(base_url, 'species', species_id,
                              'schemas', schema_id, 'loci',
                              locus_id, 'update')

        if alleles_data[i] == alleles_data[-1]:
            headers_post_bytes['complete'] = 'True'

        zip_res = cr.upload_file(current_zip, os.path.basename(current_zip),
                                 zip_url, headers_post_bytes,
                                 False)

        # determine if upload was successful
        zip_status = zip_res.status_code

        # determine if upload was successful
        if length_status not in [200, 201] or zip_status not in [200, 201]:
            failed.append(locus_id)
        elif length_status in [200, 201] and zip_status in [200, 201]:
            uploaded += 1
            print('\r', '    Sent data for alleles of '
                  '{0}/{1} loci.'.format(uploaded, len(alleles_data)), end='')

    return [failed, zip_res]


def pickle_to_fasta(locus, pickled_file, temp_dir, identifiers, reassigned):
    """Create FASTA files from pickled files.

    Parameters
    ----------
    locus : str
        The identifier of the locus with '.fasta' suffix.
    pickled_file : str
        Path to the pickled file with a dictionary that
        has integer identifiers as keys and a tuple with
        two elements: the identifier that should be assigned
        to the allele (might differ from the key if the allele
        is new, in which case it starts with '*') and the DNA
        sequence of the allele.
    temp_dir : str
        Path to the directory where the output FASTA file will
        be created.
    identifiers : dict
        The `zip_res` variable returned by the
        :py:func:`upload_alleles_data` function. It will be used
        to change allele identifiers that were successfully
        inserted into the Chewie-NS.

    Returns
    -------
    fasta_path : str
        Path to the FASTA file created by this function
    """
    locus_id = locus.rstrip('.fasta')
    locus_int = locus_id.split('-')[-1].lstrip('0')
    if locus_int in identifiers:
        repeated = identifiers[locus_int][0]
        attributed = identifiers[locus_int][1]
    else:
        repeated = {}
        attributed = {}

    inv_reassigned = {}
    if locus in reassigned:
        inv_reassigned = {v: k for k, v in reassigned[locus].items()}

    locus_sequences = fo.pickle_loader(pickled_file)

    natsorted_locus = sorted(locus_sequences)

    fasta_path = fo.join_paths(temp_dir, [locus])
    records = []
    for seqid in natsorted_locus:
        recid = locus_sequences[seqid][0]
        seq = locus_sequences[seqid][1]
        seq_hash = im.hash_sequence(seq)
        # switch by the identifier attributed by the Chewie-NS
        if seq_hash in attributed:
            new_recid = attributed[seq_hash]
            if recid in inv_reassigned:
                old_id = inv_reassigned[recid]
                reassigned[locus][old_id] = new_recid
            else:
                if locus not in reassigned:
                    reassigned[locus] = {recid: new_recid}
                else:
                    reassigned[locus][recid] = new_recid
            recid = new_recid
        elif seq_hash in repeated:
            new_recid = repeated[seq_hash]
            if recid in inv_reassigned:
                old_id = inv_reassigned[recid]
                reassigned[locus][old_id] = new_recid
            else:
                if locus not in reassigned:
                    reassigned[locus] = {recid: new_recid}
                else:
                    reassigned[locus][recid] = new_recid
            recid = new_recid

        record = '>{0}_{1}\n{2}'.format(locus_id, recid, seq)
        records.append(record)

    fasta_text = '\n'.join(records)

    with open(fasta_path, 'w') as fp:
        fp.write(fasta_text)

    os.remove(pickled_file)

    return reassigned


def retrieve_alleles(loci_new_alleles, server_time, schema_uri,
                     count, headers_get, ns_date):
    """Retrieve alleles added to Chewie-NS since last sync date.

    Retrieves alleles added to a schema in Chewie-NS
    during a time interval, up to the maximum number of
    alleles that the server returns at a time (50000).

    Parameters
    ----------
    loci_new_alleles : dict
        A dictionary with loci identifiers as keys and
        dictionaries with alleles identifiers and DNA
        sequences as values.
    server_time : str
        The function will return alleles added to the
        schema after this date (format Y-%m-%dT%H:%M:%S).
    schema_uri : str
        The URI of the schema in the Chewie-NS.
    count : int
        The cumulative number of sequences that has been
        returned.
    headers_get : dict
        HTTP headers for GET requests.
    ns_date : str
        The function will return alleles added to the
        schema up to this date (format Y-%m-%dT%H:%M:%S).

    Returns
    -------
    A list with the following variables:

    loci_new_alleles :  dict
        Input `loci_new_alleles` dictionary with alleles
        returned in the current and previous iterations.
    server_time : str
        The date of insertion of the last allele returned by
        the Chewie-NS.
    count : int
        The cumulative number of sequences that has been
        returned.
    """
    # request the new alleles starting on the date given
    url = cr.make_url(schema_uri, 'loci')
    payload = {'local_date': server_time, 'ns_date': ns_date}
    # get the new alleles
    response = cr.simple_get_request(url, headers_get, parameters=payload)[1]
    response_content = response.json()

    # get info about sequences that were added since last date
    new_alleles = response_content['newAlleles']

    # get headers info
    response_headers = response.headers
    if len(new_alleles) > 0:
        # get date of last added allele
        server_time = response_headers['Last-Allele']
        # group retrieved alleles by locus
        for allele in new_alleles:
            locus = '{0}{1}'.format(allele['locus_name']['value'], '.fasta')
            allele_id = allele['allele_id']['value']
            sequence = allele['nucSeq']['value']

            loci_new_alleles[locus][allele_id] = sequence

        # keep count of the number of retrieved alleles
        count += len(new_alleles)
    else:
        # get current server date if no alleles were added
        # since last server time
        server_time = response_headers['Server-Date']

    return (loci_new_alleles, server_time, count)


def altered_loci(loci, schema_dir, pickled_loci, not_in_ns,
                 temp_dir, rearranged):
    """Reassign local allele identifiers based on Chewie-NS nomenclature.

    Reassigns allele identifiers of loci that were altered
    in the Chewie-NS since last sync process. Identifier
    reassignment will ensure that local and remote alleles
    have the same identifiers or that alleles that are only
    local are assigned an identifier that does not match any
    identifier in the Chewie-NS.

    Parameters
    ----------
    loci : dict
        A dictionary with loci identifiers as keys and nested
        dictionaries as values (one nested dictionary per entry,
        with alleles identifiers as keys and DNA sequences as values).
    schema_dir : str
        Path to the schema's diretory.
    pickled_loci : dict
        A dictionary that will be used to store paths to
        pickled files with the records for each locus that is
        processed.
    not_in_ns : dict
        A dictionary that will be used to store information
        about local alleles that are not in the Chewie-NS.
    temp_dir : str
        Path to the directory where the output FASTA file will
        be created.

    Returns
    -------
    A list with three elements:

    - pickled_loci : dict
        Input `pickled_loci` variable updated with the paths
        to the pickled files with the records of the loci that
        were altered in the Chewie-NS since last sync process.
    - not_in_ns : dict
        Input `not_in_ns` variable that was updated with the
        information about all alleles that are not in the
        Chewie-NS.
    - rearranged : dict
        A dictionary with loci identifiers as keys and lists
        as values. Each list contains one or more two-element
        tuples that represent identifiers reassignments
        (e.g.: (*2, 3), *2 --> 3 ).
    """
    for locus, alleles in loci.items():
        locus_id = locus.rstrip('.fasta')
        # paths for current and temp locus file
        locus_file = fo.join_paths(schema_dir, [locus])

        # get the latest alleles retrieved from Chewie-NS
        ns_seqs = {seq: seqid for seqid, seq in alleles.items()}

        local_seqs = {str(rec.seq): [rec.id, (rec.id).split('_')[-1]]
                      for rec in SeqIO.parse(locus_file, 'fasta')}

        # create dictionary with updated records
        # add local records without '*'
        updated_records = {v[0]: [v[1], k] for k, v in local_seqs.items() if '*' not in v[1]}

        # add records that were added to the remote schema since last sync
        switched = {}
        for seq, seqid in ns_seqs.items():
            updated_records[seqid] = [seqid.split('_')[-1], seq]
            if seq in local_seqs:
                switched[local_seqs[seq][1]] = seqid
                # remove sequence from local sequences to avoid sending
                # duplicated sequences to Chewie-NS
                del local_seqs[seq]

        # identify largest allele identifier in Chewie-NS
        max_id = max([int(v[0]) for k, v in updated_records.items()])

        local_ids = [(v[1], k) for k, v in local_seqs.items() if '*' in v[0]]
        sorted_novel = sorted(local_ids, key=lambda x: int(x[0][1:]))

        for si in sorted_novel:
            max_id += 1
            new_id = '*{0}'.format(max_id)
            new_seqid = '{0}_{1}'.format(locus_id, new_id)
            updated_records[new_seqid] = [new_id, si[1]]
            switched[si[0]] = new_id

        if len(updated_records) > 0:
            novel_alleles = {im.hash_sequence(v[1]):
                             [v[0], v[1], len(v[1])]
                             for k, v in updated_records.items() if '*' in v[0]}
            not_in_ns[locus] = [novel_alleles]

        if len(switched) > 0:
            rearranged[locus] = switched

        updated_records2 = {}
        for seqid, v in updated_records.items():
            int_seqid = int(v[0]) if '*' not in v[0] else int(v[0][1:])
            updated_records2[int_seqid] = (seqid, v[1])

        temp_file = fo.join_paths(temp_dir, ['{0}_pickled'.format(locus_id)])
        fo.pickle_dumper(updated_records2, temp_file)

        pickled_loci[locus] = temp_file

    return [pickled_loci, not_in_ns, rearranged]


def unaltered_loci(loci, schema_dir, pickled_loci, not_in_ns, temp_dir):
    """Determine set of loci that were altered locally only.

    Determines if local schema has new alleles for loci that
    were not updated in the Chewie-NS. Creates a pickled
    file for each locus that has new alleles.

    Parameters
    ----------
    loci : dict
        A dictionary with loci identifiers as keys and nested
        dictionaries as values (one nested dictionary per entry,
        with alleles identifiers as keys and DNA sequences as values).
    schema_dir : str
        Path to the schema's diretory.
    pickled_loci : dict
        A dictionary that will be used to store paths to
        pickled files with the records for each locus that is
        processed.
    not_in_ns : dict
        A dictionary that will be used to store information
        about local alleles that are not in the Chewie-NS.
    temp_dir : str
        Path to the directory where the output FASTA file will
        be created.

    Returns
    -------
    A list with two elements:

    - pickled_loci : dict
        Input `pickled_loci` variable updated with the paths
        to the pickled files with the records of the loci that
        were not altered in the Chewie-NS since last sync process
        but were altered locally.
    - not_in_ns : dict
        Input `not_in_ns` variable that was updated with the
        information about all alleles that are not in the
        Chewie-NS.
    """
    for gene in loci:
        locus_id = gene.rstrip('.fasta')
        locus_file = fo.join_paths(schema_dir, [gene])
        # get local locus sequences
        records = {rec.id: [(rec.id).split('_')[-1], str(rec.seq)]
                   for rec in SeqIO.parse(locus_file, 'fasta')}

        # determine if there are sequences with '*'
        novel_local = {im.hash_sequence(v[1]):
                       [v[0], v[1], len(v[1])]
                       for k, v in records.items() if '*' in v[0]}
        if len(novel_local) > 0:
            not_in_ns[gene] = [novel_local]
            updated_records = {}
            for seqid, seq in records.items():
                int_seqid = int(seq[0].replace('*', ''))
                updated_records[int_seqid] = (seq[0], seq[1])

            temp_file = fo.join_paths(temp_dir, ['{0}_pickled'.format(locus_id)])
            fo.pickle_dumper(updated_records, temp_file)

            pickled_loci[gene] = temp_file

    return [pickled_loci, not_in_ns]


def update_loci_files(loci, local_loci, schema_dir, temp_dir):
    """Determine set loci/alleles to update in the local and remote schemas.

    Determines which loci were or were not updated in the
    remote schema, alleles that are not in the Chewie-NS,
    reassigns identifiers to ensure that alleles that are
    common to local and remote schema have the same
    identifiers and saves updated records to pickled files
    to allow the update of the local schema and of the
    remote schema by sending novel alleles to the Chewie-NS.

    Parameters
    ----------
    loci : dict
        Dictionary with the alleles added to the remote
        schema since last sync date.
    local_loci : list
        List with the names of the schema's loci files.
    schema_dir : str
        Path to the schema's diretory.
    temp_dir : str
        Path to the directory where the output pickled files
        will be created.

    Returns
    -------
    A list with five elements:

    - not_in_ns : dict
        Dictionary with information about local alleles that are not
        in the Chewie-NS. It has loci identifiers as keys and lists
        as values. Each list contains the following elements:

        - A dictionary with sequences hashes as keys and a list
          with the sequence identifier, DNA sequence and sequence
          length value.
        - The locus URI.
    - pickled_loci : dict
        A dictionary with paths to pickled files that contain the
        records for each locus that has been altered in the remote
        schema and/or locally with new alleles added during the
        AlleleCall process.
    - updated : dict
        Dictionary with the alleles added to the remote
        schema since last sync date.
    - not_updated : list
        List of loci that were not altered in the remote schema
        since last sync process.
    - rearranged : dict
        A dictionary with loci identifiers as keys and lists
        as values. Each list contains one or more two-element
        tuples that represent identifiers reassignments
        (e.g.: (*2, 3), *2 --> 3 ).
    """
    rearranged = {}
    not_in_ns = {}
    pickled_loci = {}
    # loci in the remote schema that were altered
    updated = {gene: loci[gene] for gene in local_loci if gene in loci}
    # loci that were not altered in the remote schema
    not_updated = [gene for gene in local_loci if gene not in loci]

    # update local identifiers based on identifiers in the Chewie-NS
    pickled_loci, not_in_ns, \
        rearranged = altered_loci(updated, schema_dir, pickled_loci,
                                  not_in_ns, temp_dir, rearranged)

    # determine loci that were not altered in the Chewie-NS but
    # have been altered locally
    pickled_loci, not_in_ns = unaltered_loci(not_updated, schema_dir,
                                             pickled_loci, not_in_ns,
                                             temp_dir)

    return [not_in_ns, pickled_loci, rearranged]


def retrieve_latest(local_date, schema_uri, headers_get, ns_date):
    """Retrieve alleles added to a schema in Chewie-NS after a specified date.

    Parameters
    ----------
    local_date : str
        Last sync date. The function will return alleles added
        to the remote schema after this date (format Y-%m-%dT%H:%M:%S).
    schema_uri : str
        The URI of the schema in the Chewie-NS.
    headers_get : dict
        HTTP headers for GET requests.
    ns_date : str
        Last modification date of the remote schema in the Chewie-NS.
        The function will return alleles added to the schema up to
        this date (format Y-%m-%dT%H:%M:%S).

    Returns
    -------
    A list with the following variables:

    new_alleles : dict
        A dictionary with loci identifiers as keys and
        dictionaries with alleles identifiers and DNA
        sequences as values.
    server_time : str
        The last modification date of the remote schema.
    count : int
        The total number of sequences that were retrieved from
        the Chewie-NS.
    """
    # get list of sequences that are new considering the last date
    count = 0
    server_time = local_date
    new_alleles = defaultdict(dict)

    # get new alleles from Chewie-NS
    while server_time != ns_date:
        new_alleles, server_time, count = retrieve_alleles(new_alleles,
                                                           server_time,
                                                           schema_uri,
                                                           count,
                                                           headers_get,
                                                           ns_date)

    return [new_alleles, server_time, count]


def main(schema_directory, cpu_cores, nomenclature_server,
         submit, blast_path):  # update_profiles):

    # get ns configs
    local_date, schema_uri = pv.read_configs(schema_directory, '.ns_config')
    # get schema and species identifiers
    schema_id = schema_uri.split('/')[-1]
    species_id = schema_uri.split('/')[-3]
    if nomenclature_server is None:
        nomenclature_server = schema_uri.split('species/')[0]

    if submit is True and 'tutorial' not in nomenclature_server:
        print('\nOnly authorized registered users may submit new alleles.')
        token = cr.capture_login_credentials(nomenclature_server)
    else:
        token = ''

    # GET request headers
    headers_get = ct.HEADERS_GET_JSON
    headers_get['Authorization'] = token

    # determine current user ID and Role
    if submit is True and 'tutorial' not in nomenclature_server:
        user_id, user_role, user_auth = cr.user_info(nomenclature_server, headers_get)
        # verify if user has authorization to submit
        url = cr.make_url(nomenclature_server, 'auth', 'check')
        response = cr.simple_get_request(url, headers_get)[1]
        if response.status_code == 200:
            user_auth = True
        else:
            sys.exit('Current user has no authorization to submit novel '
                     'alleles.\nYou can request authorization to submit '
                     'novel alleles by sending an e-mail to: '
                     'imm-bioinfo@medicina.ulisboa.pt')
        print('User id: {0}'.format(user_id))
        print('User role: {0}\n'.format(user_role))
    else:
        user_id = ''
        user_role = ''
        user_auth = True if 'tutorial' in nomenclature_server else False

    # POST requests headers
    headers_post = ct.HEADERS_POST_JSON
    headers_post['Authorization'] = token
    headers_post['user_id'] = user_id
    # POST headers to send binary data
    headers_post_bytes = ct.HEADERS_POST
    headers_post_bytes['Authorization'] = token
    headers_post_bytes['user_id'] = user_id

    schema_params = pv.read_configs(schema_directory, '.schema_config')

    # verify that local configs have a single value per parameter
    if all([len(schema_params[k]) == 1
            for k in schema_params
            if k != 'chewBBACA_version']) is not True:
        sys.exit('Cannot sync schema with multiple values per parameter.')

    # check if schema exists in Chewie-NS
    schema_name, ns_params = cr.get_species_schemas(schema_id,
                                                    species_id,
                                                    nomenclature_server,
                                                    headers_get)[2:]

    # verify that local configs match NS configs
    if all([str(schema_params[k][0]) == ns_params[k]['value']
            for k in schema_params
            if k not in ['chewBBACA_version', 'window_size']]) is not True:
        sys.exit('Local configs do not match Chewie-NS configs.')

    # Get the name of the species from the provided id
    # or vice-versa
    species_id, species_name = cr.species_ids(species_id, nomenclature_server,
                                              headers_get)

    print('Schema id: {0}'.format(schema_id))
    print('Schema name: {0}'.format(schema_name))
    print("Schema's species: {0} (id={1})".format(species_name, species_id))
    print('Last synced: {0}'.format(local_date))

    # get last modification date
    # setting syncing date to last modification date will allow
    # all users to sync even when the schema is locked and being
    # updated by another user
    ns_date = ns_params['last_modified']['value']
    print('\nRemote schema was last modified on: {0}'.format(ns_date))

    # exit if remote schema has not been updated since last
    # sync date and current user does not wish to submit new alleles
    if local_date == ns_date and submit is False:
        sys.exit('\nRemote schema has not been updated since last sync '
                 'process. Local schema is up-to-date.')

    # Create a temporary dir for the new alleles
    temp_dir = fo.join_paths(os.path.dirname(schema_directory), ['temp'])
    if not os.path.exists(temp_dir):
        os.mkdir(temp_dir)

    # retrieve alleles added to schema after last sync date
    print('\nRetrieving alleles added to remote schema '
          'after {0}...'.format(local_date))
    loci_alleles, server_time, count = retrieve_latest(local_date, schema_uri,
                                                       headers_get, ns_date)

    print('Retrieved {0} alleles for {1} loci.'
          ''.format(count, len(loci_alleles)))

    # Get schema files from genes list file
    genes_list = fo.join_paths(schema_directory, ['.genes_list'])
    genes = fo.pickle_loader(genes_list)

    # update loci structure
    not_in_ns, pickled_loci, rearranged = update_loci_files(loci_alleles, genes,
                                                            schema_directory, temp_dir)

    total_local = sum([len(v[0]) for k, v in not_in_ns.items()])
    print('Local schema has {0} novel alleles for {1} '
          'loci.'.format(total_local, len(not_in_ns)))

    # check if there are any changes to make
    if len(pickled_loci) == 0:
        shutil.rmtree(temp_dir)
        sys.exit('Remote schema has not been altered and local schema '
                 'does not have novel alleles.')

    results = {}
    attributed = 0
    if submit is True and user_auth is True and len(not_in_ns) > 0:

        # attempt to lock schema
        lock_res = cr.simple_post_request(nomenclature_server, headers_post,
                                          ['species', species_id,
                                           'schemas', schema_id,
                                           'lock'],
                                          data=json.dumps({'action': 'lock'}))[1]
        # if schema is already locked user cannot send alleles
        lock_status = lock_res.status_code
        if lock_status == 403:
            print('Schema is already locked. Another user might be updating '
                  'the schema. Please repeat the syncing process after a '
                  'while to add your new alleles to the Chewie-NS.\n The '
                  'process will now update your local schema with the alleles '
                  'retrieved from the Chewie-NS.')
        else:

            # after locking, check if date matches ns_date
            date_res = cr.simple_get_request(nomenclature_server, headers_get,
                                             ['species', species_id, 'schemas',
                                              schema_id, 'modified'])[1]

            date_value = (date_res.json()).split(' ')[-1]

            if date_value != ns_date:
                print('Data retrieved from the Chewie-NS has an older '
                      'timestamp than current schema timestamp. Schema '
                      'might have been updated before this syncing process. '
                      'Please repeat the syncing process in order to add '
                      'your new alleles to the schema. The process will now '
                      'update your local schema with the alleles retrieved '
                      'from the Chewie-NS.')

                # unlock schema
                lock_res = cr.simple_post_request(nomenclature_server, headers_post,
                                                  ['species', species_id, 'schemas',
                                                   schema_id, 'lock'],
                                                  data=json.dumps({'action': 'unlock'}))[1]
            else:
                print('Collecting data and creating files to submit local alleles...')
                # get list of loci for schema in the NS
                loci_res = cr.simple_get_request(nomenclature_server, headers_get,
                                                 ['species', species_id, 'schemas',
                                                  schema_id, 'loci'])[1]
                # get loci files names from response
                for l in loci_res.json()['Loci']:
                    locus_name = l['name']['value'] + '.fasta'
                    locus_uri = l['locus']['value']
                    if locus_name in not_in_ns:
                        not_in_ns[locus_name].append(locus_uri)

                # create files with length values to update
                length_files = create_lengths_files(not_in_ns, temp_dir)

                # create new alleles data
                alleles_files, \
                    loci_ids, \
                    loci_names = create_alleles_files(not_in_ns, nomenclature_server,
                                                      user_id, species_name,
                                                      species_id, schema_id,
                                                      temp_dir)

                # compress files with new alleles
                zipped_files = ['{0}.zip'.format(file) for file in alleles_files]
                list(map(fo.file_zipper, alleles_files, zipped_files))
                alleles_data = list(zip(zipped_files, loci_ids, loci_names))

                print('Sending and inserting new alleles...')
                failed, \
                    start_count = upload_alleles_data(alleles_data, length_files,
                                                      nomenclature_server, headers_post,
                                                      headers_post_bytes, species_id,
                                                      schema_id)

                # track progress through endpoint
                # set time limit for task completion (seconds)
                print()
                time_limit = 2100
                current_time = 0
                status = 'Updating'
                start_count = int(start_count.json()['nr_alleles'])
                while status != 'Complete' and (current_time < time_limit):
                    insertion_status = cr.simple_get_request(nomenclature_server,
                                                             headers_get,
                                                             ['species', species_id,
                                                              'schemas', schema_id,
                                                              'loci', 'locus', 'update'])[1]
                    insertion_status = insertion_status.json()
                    if 'message' in insertion_status:
                        status = 'Complete'
                        results = insertion_status['identifiers']

                    current_count = int(insertion_status['nr_alleles'])

                    inserted = current_count - start_count
                    print('\r', '    Inserted {0} alleles.'.format(inserted), end='')
                    time.sleep(2)
                    current_time += 2

                if current_time != time_limit:
                    # determine alleles that were attributed an identifier
                    repeated = sum([len(r[0]) for l, r in results.items()])
                    attributed = sum([len(r[1]) for l, r in results.items()])

                    print('\nThe Chewie-NS inserted {0} new alleles and detected '
                          '{1} repeated alleles.'.format(attributed, repeated))
                else:
                    print('\nCould not retrieve allele identifiers assigned by '
                          'Chewie-NS. Will adapt schema with retrieved alleles. '
                          'Please repeat the syncing process in order to assign '
                          'the new identifiers for the submitted alleles.')

                # remove files in temp folder
                fo.remove_files(length_files)
                fo.remove_files(alleles_files)
                fo.remove_files(zipped_files)

    # change pickled files to FASTA files
    for locus, pick in pickled_loci.items():
        rearranged = pickle_to_fasta(locus, pick, temp_dir, results,
                                     rearranged)

    # update pre-computed tables
    # pre_computed_dir = fo.join_paths(schema_directory, ['pre_computed'])
    # if os.path.isdir(pre_computed_dir):
    #     schema_loci_fullpath = [fo.join_paths(schema_directory, [file])
    #                             for file in genes]
    #     loci_to_call = {file: schema_loci_fullpath.index(file)
    #                     for file in schema_loci_fullpath}
    #     update_hash_tables(rearranged, loci_files, loci_to_call,
    #                        schema_params['translation_table'][0],
    #                        pre_computed_dir)
    # delete pre-computed directory
    # future versions will change the pre-computed hash tables
    pre_computed_dir = fo.join_paths(schema_directory, ['pre_computed'])
    if os.path.isdir(pre_computed_dir):
        fo.delete_directory(pre_computed_dir)

    # change identifiers in SQLite DB
    # if len(rearranged) > 0 and update_profiles is True:
    #     print('\nUpdating local allele identifiers...')
    #     altered = ps.update_profiles(schema_directory, rearranged)
    #     if altered is not None:
    #         print('Updated {0} profiles.\n'.format(altered))
    #     else:
    #         print('Could not find local SQLite database to upload profiles.\n')

    # Re-determine the representative sequences
    if attributed > 0 or count > 0:
        PrepExternalSchema.main(temp_dir, schema_directory,
                                cpu_cores, float(schema_params['bsr'][0]),
                                int(schema_params['minimum_locus_length'][0]),
                                11, '', None, blast_path)

        # delete invalid alleles and genes files
        parent_dir = os.path.dirname(schema_directory)
        files = [fo.join_paths(parent_dir, [file])
                 for file in os.listdir(parent_dir)
                 if 'invalid' in file]

        fo.remove_files(files)

        # get last modification date
        last_modified = cr.simple_get_request(nomenclature_server, headers_get,
                                              ['species', species_id,
                                               'schemas', schema_id,
                                               'modified'])[1]
        last_modified = (last_modified.json()).split(' ')[-1]
        server_time = last_modified

        # update NS config file with latest server time
        ns_configs = fo.join_paths(schema_directory, ['.ns_config'])
        fo.pickle_dumper([server_time, schema_uri], ns_configs)

    print('Received {0} new alleles for {1} loci and sent '
          '{2} for {3} loci. '.format(count, len(pickled_loci),
                                      attributed, len(not_in_ns)))

    # delete temp directory
    shutil.rmtree(temp_dir)

    # delete pre-computed loci mode values
    mode_file = fo.join_paths(schema_directory, ['loci_modes'])
    fo.remove_files([mode_file])

    # delete pre-computed BSR values from 'short' directory
    # representatives might have changed and BSR values are outdated
    bsr_file = fo.join_paths(schema_directory, ['short', 'self_scores'])
    fo.remove_files([bsr_file])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Purpose
-------

This module contains functions related to

Code documentation
------------------
"""


import os
import sys

try:
    from utils import (constants as ct,
                       blast_wrapper as bw,
                       gene_prediction as gp,
                       file_operations as fo,
                       fasta_operations as fao,
                       sequence_clustering as sc,
                       sequence_manipulation as sm,
                       iterables_manipulation as im,
                       multiprocessing_operations as mo)
except ModuleNotFoundError:
    from CHEWBBACA.utils import (constants as ct,
                                 blast_wrapper as bw,
                                 gene_prediction as gp,
                                 file_operations as fo,
                                 fasta_operations as fao,
                                 sequence_clustering as sc,
                                 sequence_manipulation as sm,
                                 iterables_manipulation as im,
                                 multiprocessing_operations as mo)


def predict_genes(fasta_files, ptf_path, translation_table,
                  prodigal_mode, cpu_cores, output_directory):
    """Execute Prodigal to predict coding sequences from Fasta files.

    Parameters
    ----------
    fasta_files : list
        List of paths to FASTA files with genomic
        sequences.
    ptf_path : str
        Path to the Prodigal training file. Should
        be NoneType if a training file is not provided.
    translation_table : int
        Genetic code used to predict and translate
        coding sequences.
    prodigal_mode : str
        Prodigal execution mode.
    cpu_cores : int
        Number of processes that will run Prodigal in
        parallel.
    output_directory : str
        Path to the directory where output files
        with Prodigal's results will be stored in.

    Returns
    -------
    failed_info : list
        List that contains a list with the stderr for the
        cases that Prodigal failed to predict genes for
        and the path to the file with information about
        failed cases. Returns NoneType if gene prediction
        succeeded for all inputs.
    """
    # divide input genomes into equal number of sublists for
    # maximum process progress resolution
    prodigal_inputs = im.divide_list_into_n_chunks(fasta_files,
                                                   len(fasta_files))

    common_args = [output_directory, ptf_path,
                   translation_table, prodigal_mode]

    # add common arguments to all sublists
    prodigal_inputs = im.multiprocessing_inputs(prodigal_inputs,
                                                common_args,
                                                gp.main)

    # run Prodigal to predict genes
    prodigal_results = mo.map_async_parallelizer(prodigal_inputs,
                                                 mo.function_helper,
                                                 cpu_cores,
                                                 show_progress=True)

    # determine if Prodigal predicted genes for all genomes
    failed = {line[0]: line[1]
              for line in prodigal_results
              if line[1] == 0
              or isinstance(line[1], str) is True}

    return failed


def extract_genes(fasta_files, prodigal_path, cpu_cores,
                  temp_directory):
    """Extract coding sequences from FASTA files.

    Extract coding sequences from genomic Fasta files and
    saves coding sequences and info about coding sequences.

    Parameters
    ----------
    fasta_files : list
        List of paths to FASTA files with genomic sequences.
    prodigal_path : str
        Path to the directory with the files with Prodigal
        results.
    cpu_cores : int
        Number of processes that will extract coding sequences
        in parallel.
    temp_directory : str
        Path to the directory where FASTA files with extracted
        coding sequences will be stored in.

    Returns
    -------
    cds_files : list
        List with paths to the FASTA files that contain
        extracted coding sequences.
    total_extracted : int
        Total number of coding sequences extracted from the
        input fasta files.
    """
    # divide inputs into 15 sublists for ~7% process progress resolution
    num_chunks = 15
    extractor_inputs = im.divide_list_into_n_chunks(fasta_files, num_chunks)

    # add common arguments and unique index/identifier
    output_index = [i+1 for i in range(len(extractor_inputs))]
    extractor_inputs = im.aggregate_iterables([extractor_inputs, output_index])
    extractor_inputs = im.multiprocessing_inputs(extractor_inputs,
                                                 [prodigal_path, temp_directory],
                                                 gp.cds_batch_extractor)

    # extract coding sequences
    extracted_cdss = mo.map_async_parallelizer(extractor_inputs,
                                               mo.function_helper,
                                               cpu_cores,
                                               show_progress=True)

    total_extracted = sum([f[2] for f in extracted_cdss])

    # create full table file
    table_files = [f[0] for f in extracted_cdss]
    table_file = fo.join_paths(temp_directory, [ct.CDS_COORDINATES_BASENAME])
    fo.concatenate_files(table_files, table_file, ct.CDS_TABLE_HEADER)
    fo.remove_files(table_files)

    cds_files = [f[1] for f in extracted_cdss]

    return [cds_files, total_extracted, table_file]


def exclude_duplicates(fasta_files, temp_directory, cpu_cores,
                       outfile_template, ids_map, protein=False,
                       only_seqids=False):
    """Identify duplicated sequences and select distinct set of sequences.

    Parameters
    ----------
    fasta_files : list
        List with paths to FASTA files.
    temp_directory : str
        Path to the directory where new files will be
        created.
    cpu_cores : int
        Number of deduplication processes to run in
        parallel.
    outfile_template : str
        Template for the name of output files.

    Returns
    -------
    A list with the following elements:
        distinct_seqids : list
            List with the sequence identifiers of distinct
            sequences.
        distinct_seqs : str
            Path to the FASTA file with distinct sequences.
    """
    # create groups of inputs for multiprocessing
    output_files = [fo.join_paths(temp_directory,
                                  [outfile_template.format(str(i+1)+'.fasta')])
                    for i, file in enumerate(fasta_files)]

    inputs = im.aggregate_iterables([fasta_files, output_files])

    dedup_inputs = im.multiprocessing_inputs(inputs,
                                             [ids_map[0]],
                                             sm.determine_distinct)

    # determine distinct sequences
    dedup_results = mo.map_async_parallelizer(dedup_inputs,
                                              mo.function_helper,
                                              cpu_cores,
                                              show_progress=False)

    repeated = 0
    merged_results = {}
    distinct_seqids = []
    # merge results
    for p in dedup_results:
        r = fo.pickle_loader(p)
        for k, v in r.items():
            if k in merged_results:
                stored_ids = im.polyline_decoding(merged_results[k])
                if protein is False:
                    rest = [v[i] for i in range(1, len(v), 2)]
                    merged_results[k] = im.polyline_encoding(stored_ids+rest)
                else:
                    merged_results[k] = im.polyline_encoding(stored_ids+v)
                repeated += (len(v)/2)
            else:
                seqid = '{0}-protein{1}'.format(ids_map[1][v[1]], v[0])
                distinct_seqids.append(seqid)
                if protein is False:
                    rep = v[0:2]
                    rest = [v[i] for i in range(3, len(v), 2)]
                    merged_results[k] = im.polyline_encoding(rep+rest)
                else:
                    merged_results[k] = im.polyline_encoding(v)

                repeated += (len(v)/2) - 1

    # save table with deduplicated records
    hash_table_file = fo.join_paths(temp_directory, [outfile_template.format('merged.hashtable')])
    fo.pickle_dumper(merged_results, hash_table_file)

    # remove intermediate deduplication tables
    fo.remove_files(dedup_results)

    # concatenate Fasta files from parallel processes
    dedup_files = [f[1] for f in dedup_inputs]
    cds_file = fo.join_paths(temp_directory, [outfile_template.format('concat.fasta')])
    cds_file = fo.concatenate_files(dedup_files, cds_file)

    # create index for concatenated Fasta
    cds_index = fao.index_fasta(cds_file)

    # define filename for file with distinct sequences
    distinct_seqs = fo.join_paths(temp_directory,
                                  [outfile_template.format('merged.fasta')])
    # get the representative record for each distinct sequence
    fao.get_sequences_by_id(cds_index, distinct_seqids,
                            distinct_seqs, 50000)

    fo.remove_files(dedup_files+[cds_file])

    if only_seqids is False:
        return [merged_results, distinct_seqs, repeated]
    else:
        return [distinct_seqids, distinct_seqs, repeated]


def exclude_small(fasta_file, minimum_length, variation=0):
    """Identify sequences smaller that a specified length value.

    Parameters
    ----------
    fasta_file : str
        Path to a FASTA file.
    minimum_length : int
        Sequences with a length value below this value are
        considered small.
    variation : float
        Accept sequences with length variation of up to
        minus (`minimum_length`*`variation`).

    Returns
    -------
    small_seqids : list
        List with the sequence identifiers of small
        sequences.
    ss_lines : list
        List with one string per small sequence. Each string
        represents an exception message for a sequence that
        is small.
    """
    # determine small sequences and keep their seqids
    small_seqids = sm.determine_small(fasta_file, minimum_length, variation)

    ss_lines = ['{0}: smaller than {1} chars'.format(seqid, minimum_length)
                for seqid in small_seqids]

    return [small_seqids, ss_lines]


def translate_sequences(sequence_ids, sequences_file, temp_directory,
                        translation_table, minimum_length, cpu_cores):
    """Translate DNA sequences.

    Returns information about sequences that are untranslatable
    and saves translatable DNA sequences and proteins to FASTA files.

    Parameters
    ----------
    sequence_ids : list
        List with the identifiers of the sequences that
        should be translated.
    sequences_file : str
        Path to the FASTA file with the DNA sequences.
    temp_directory : str
        Path to the directory where new files will be
        created.
    translation_table : int
        Genetic code used to translate coding sequences.
    minimum_length : int
        Sequences with a length value below this value are
        considered small and are excluded.
    cpu_cores : int
        Number of translation processes to run in
        parallel.

    Returns
    -------
    A list with the following elements:
        dna_file : str
            Path to the FASTA with all DNA sequences
            that could be translated.
        protein_file : str
            Path to the FASTA file with the translated
            sequences.
        untrans_seqids : list
            List with the identifiers of the sequences
            that could not be translated.
        untrans_lines : list
            List with one string per untranslatable
            sequence. Each string has a sequence
            identifier and a small description that
            indicates why the sequence could not be
            translated.
    """
    # divide inputs into sublists
    translation_inputs = im.divide_list_into_n_chunks(sequence_ids, cpu_cores)

    # create paths to files with protein sequences
    protein_template = fo.join_paths(temp_directory, ['translated_cds_{0}.fasta'])
    protein_files = [protein_template.format(i+1)
                     for i in range(len(translation_inputs))]

    # add common args to sublists
    common_args = [sequences_file, translation_table, minimum_length]
    translation_inputs = im.aggregate_iterables([translation_inputs,
                                                 protein_files])
    translation_inputs = im.multiprocessing_inputs(translation_inputs,
                                                   common_args,
                                                   sm.translate_coding_sequences)

    # translate sequences
    translation_results = mo.map_async_parallelizer(translation_inputs,
                                                    mo.function_helper,
                                                    cpu_cores,
                                                    show_progress=True)

    # concatenate files
    protein_file = fo.join_paths(temp_directory, ['translated_cds_concat.fasta'])
    protein_file = fo.concatenate_files(protein_files, protein_file)
    fo.remove_files(protein_files)

    # determine sequences that could not be translated
    untrans_lines = []
    untrans_seqids = []
    for res in translation_results:
        if len(res[0]) > 0:
            untrans_lines.extend(['{0}: {1}'.format(r[0], r[1])
                                  for r in res[0]])
            untrans_seqids.extend([r[0] for r in res[0]])

    return [protein_file, untrans_seqids, untrans_lines]


def cluster_sequences(sequences, word_size, window_size, clustering_sim,
                      representatives, grow_clusters, kmer_offset,
                      seq_num_cluster, temp_directory, cpu_cores,
                      divide, position):
    """Cluster sequences based on the proportion of shared minimizers.

    Parameters
    ----------
    sequences : dict
        Dictionary with sequence identifiers as keys and
        sequences as values.
    word_size : int
        Value k for the k-mer size.
    window_size : int
        Value for the window size/number of consecutive
        k-mers per window.
    clustering_sim : float
        Similarity threshold to add a sequence to
        a cluster.
    representatives : dict
        Dictionary with k-mers as keys and a list with
        identifiers of sequences that contain that k-mer
        as values.
    grow_clusters : bool
        If it is allowed to create new clusters.
    kmer_offset : int
        Value to indicate offset of consecutive kmers.
    seq_num_cluster : int
        Maximum number of clusters that a sequence can be
        added to.
    temp_directory : str
        Path to the directory where the clustering results
        will be saved to.
    cpu_cores : int
        Number of clustering processes to run in parallel.
    divide : bool
        If input sequences should be divided into smaller
        groups that can be processed in parallel.
    position : bool
        True if the start position for each k-mer should be saved.
        False otherwise.

    Returns
    -------
    clusters : dict
        Dictionary with representative sequence identifiers
        as keys and lists of tuples as values. Each tuple
        contains a sequence identifier of a clustered
        sequence, the proportion of shared kmers with the
        representative and the length of the clustered
        sequence.
    """
    # sort sequences by length
    sorted_seqs = {k: v for k, v in im.sort_iterable(sequences.items(),
                                                     sort_key=lambda x: len(x[1]),
                                                     reverse=True)}

    if divide is not None:
        # divide sequences into sublists
        # do not divide based on number of available cores as it may
        # lead to different results with different number of cores
        # divide into clusters with fixed number of sequences
        cluster_inputs = im.split_iterable(sorted_seqs, divide)
    else:
        cluster_inputs = [sorted_seqs]

    common_args = [word_size, window_size, clustering_sim,
                   representatives, grow_clusters, kmer_offset,
                   position, seq_num_cluster,
                   sc.clusterer]
    cluster_inputs = [[c, *common_args] for c in cluster_inputs]

    # cluster proteins in parallel
    clustering_results = mo.map_async_parallelizer(cluster_inputs,
                                                   mo.function_helper,
                                                   cpu_cores,
                                                   show_progress=True)

    # merge clusters
    clusters = [d[0] for d in clustering_results]
    clusters = im.merge_dictionaries(clusters)
    rep_sequences = [d[1] for d in clustering_results]
    rep_sequences = im.merge_dictionaries(rep_sequences)

    # perform clustering with representatives
    # this step does not run for AlleleCall
    if len(cluster_inputs) > 1 and any(len(r) > 0 for r in rep_sequences) is True:
        # cluster representatives
        rep_clusters = sc.clusterer(rep_sequences, word_size,
                                    window_size, clustering_sim,
                                    representatives, grow_clusters,
                                    kmer_offset, position,
                                    seq_num_cluster)

        merged_clusters = {}
        for k, v in rep_clusters[0].items():
            # merge clusters whose representatives are similar
            for n in v:
                # representatives from other clusters are added with
                # similarity score against new representative
                # clustered sequences from other clusters are added
                # with similarity score against their representative
                add_seqids = [n] + [s for s in clusters[n[0]] if s[0] != n[0]]
                merged_clusters.setdefault(k, []).extend(add_seqids)

        clusters = merged_clusters

    # sort clusters
    clusters = {k: v for k, v in im.sort_iterable(clusters.items())}

    # write file with clustering results
    clusters_out = os.path.join(temp_directory, 'clusters.txt')
    sc.write_clusters(clusters, clusters_out)

    return clusters


def cluster_representative_filter(clusters, representative_filter,
                                  output_directory):
    """Exclude sequences highly similar to cluster representatives.

    After removing highly similar sequences, excludes clusters
    that are singletons (only contain the representative).

    Parameters
    ----------
    clusters : dict
        Dictionary with representative sequence identifiers
        as keys and lists of tuples as values. Each tuple
        contains a sequence identifier of a clustered
        sequence, the proportion of shared kmers with the
        representative and the length of the clustered
        sequence.
    representative_filter : float
        Similarity threshold value. Sequences with
        equal or greater similarity value with the cluster's
        representative are excluded from clusters.
    output_directory : str
        Path to the directory where the clustering results
        will be saved to.

    Returns
    -------
    A list with the following elements:
        pruned_clusters : dict
            Clusters without the sequences that were highly
            similar to the cluster's representative and without
            the clusters that were singletons (only contained
            the representative).
        excluded_seqids : list
            List with the sequence identifiers of the sequences
            that were excluded from the clusters.
    """
    # remove sequences that are very similar to representatives
    pruning_results = sc.representative_pruner(clusters,
                                               representative_filter)

    pruned_clusters, excluded_seqids = pruning_results

    # get identifiers of excluded sequences
    # determine set because same seqids could be in several clusters
    excluded_seqids = set([e[0] for e in excluded_seqids])

    # remove excluded seqids from clusters without high representative
    # similarity
    pruned_clusters = {k: [e for e in v if e[0] not in excluded_seqids]
                       for k, v in pruned_clusters.items()}

    # write file with pruning results
    pruned_out = os.path.join(output_directory, 'clusters.txt')
    sc.write_clusters(pruned_clusters, pruned_out)

    # identify singletons and exclude those clusters
    singletons = im.select_keys(pruned_clusters, 0)

    pruned_clusters = im.prune_dictionary(pruned_clusters, singletons)

    # determine number of sequences that still need to be evaluated
    # +1 to include representative
    clustered_sequences = sum([len(v)+1 for k, v in pruned_clusters.items()]) + len(singletons)

    # write list of excluded seqids to file
    excluded_outfile = os.path.join(output_directory, 'excluded.txt')
    fo.write_lines(excluded_seqids, excluded_outfile)

    return [pruned_clusters, excluded_seqids, singletons, clustered_sequences]


def cluster_intra_filter(clusters, sequences, word_size,
                         intra_filter, output_directory):
    """Exclude sequences that are highly similar to other clustered sequences.

    Parameters
    ----------
    clusters : dict
        Dictionary with representative sequence identifiers
        as keys and lists of tuples as values. Each tuple
        contains a sequence identifier of a clustered
        sequence, the proportion of shared kmers with the
        representative and the length of the clustered
        sequence.
    sequences : dict
        Dictionary with sequence identifiers as keys and
        sequences as values.
    word_size : int
        Value k for the k-mer size.
    intra_filter : float
        Similarity threshold value. Sequences with
        equal or greater similarity value with other
        clustered sequences are excluded from clusters.
    output_directory : str
        Path to the directory where the clustering results
        will be saved to.

    Returns
    -------
    A list with the following elements:
        clusters : dict
            Clusters without the sequences that were highly
            similar to other clustered sequences.
        intra_excluded : list
            List with the identifiers of the sequences that
            were excluded.
    """
    # identify clusters with more than 1 sequence
    intra_clusters = {k: v for k, v in clusters.items() if len(v) > 1}

    excluded_seqids, excluded_sims = sc.intra_cluster_sim(intra_clusters,
                                                          sequences,
                                                          word_size,
                                                          intra_filter)

    intra_excluded = [v for k, v in excluded_seqids.items()]
    intra_excluded = im.flatten_list(intra_excluded)
    # get identifiers of excluded sequences
    # determine set because same seqids could be in several clusters
    intra_excluded = set(intra_excluded)

    # remove excluded seqids from clusters without high intra-similarity
    pruned_clusters = {k: [e for e in v if e[0] not in intra_excluded]
                       for k, v in clusters.items()}

    # write excluded to file
    intrasim_out = os.path.join(output_directory, 'excluded.txt')
    sc.write_clusters(excluded_sims, intrasim_out)
    # write clusters to file
    intrasim_out = os.path.join(output_directory, 'clusters.txt')
    sc.write_clusters(pruned_clusters, intrasim_out)

    return [pruned_clusters, intra_excluded]


def blast_clusters(clusters, sequences, output_directory,
                   blastp_path, makeblastdb_path, cpu_cores,
                   only_rep=False):
    """Use BLAST to align sequences in the same clusters.

    Parameters
    ----------
    clusters : dict
        Dictionary with representative sequence identifiers
        as keys and lists of tuples as values. Each tuple
        contains a sequence identifier of a clustered
        sequence, the proportion of shared kmers with the
        representative and the length of the clustered
        sequence.
    sequences : dict
        Dictionary with sequence identifiers as keys and
        sequences as values.
    output_directory : str
        Path to the directory where the clustering results
        will be saved to.
    blastp_path : str
        Path to the `BLASTp` executable.
    makeblastdb_path : str
        Path to the `makeblastdb` executable.
    cpu_cores : int
        Number of BLASTp processes to run in parallel.

    Returns
    -------
    A list with the following elements:
        blast_results : list
            List with paths to the files with BLASTp results
            (one file per cluster).
        ids_dict : dict
            Dictionary that maps sequence identifiers to
            shorter and unique integer identifiers used
            to avoid errors during BLAST execution related with
            sequence headers/identifiers that exceed the
            length limit allowed by BLAST.
    """
    # create FASTA file with sequences in clusters
    clustered_seqs_file = fo.join_paths(output_directory,
                                        ['clustered_sequences.fasta'])
    clustered_sequences = [[k]+[e[0] for e in v] for k, v in clusters.items()]
    clustered_sequences = im.flatten_list(clustered_sequences)
    # do not include duplicate identifiers
    clustered_sequences = list(set(clustered_sequences))
    fao.get_sequences_by_id(sequences, clustered_sequences,
                            clustered_seqs_file)

    # create FASTA file with replaced headers to avoid header
    # length limitation in BLAST
    integer_clusters = fo.join_paths(output_directory,
                                     ['clustered_sequences_integer_headers.fasta'])
    ids_dict = fao.integer_headers(clustered_seqs_file, integer_clusters,
                                   prefix='seq_')

    # create BLAST DB
    blast_db = fo.join_paths(output_directory,
                             ['clustered_sequences'])
    db_stderr = bw.make_blast_db(makeblastdb_path, integer_clusters,
                                 blast_db, 'prot')
    if len(db_stderr) > 0:
        sys.exit(db_stderr)

    blast_results_dir = os.path.join(output_directory,
                                     'BLAST_results')
    fo.create_directory(blast_results_dir)

    # create files with replaced sequence identifiers per cluster
    seqids_to_blast = sc.blast_seqids(clusters, blast_results_dir, ids_dict,
                                      only_rep)

    # distribute clusters per available cores
    process_num = 20 if cpu_cores <= 20 else cpu_cores
    splitted_seqids = mo.distribute_loci(seqids_to_blast, process_num, 'seqcount')

    common_args = [integer_clusters, blast_results_dir, blastp_path,
                   blast_db, only_rep, sc.cluster_blaster]

    splitted_seqids = [[s, *common_args] for s in splitted_seqids]

    # BLAST sequences in a cluster against every sequence in that cluster
    blast_results = mo.map_async_parallelizer(splitted_seqids,
                                              mo.function_helper,
                                              cpu_cores,
                                              show_progress=True)

    return [blast_results, ids_dict]


def compute_bsr(subject_score, query_score):
    """Compute the BLAST Score Ratio for an alignment between two sequences.

    Parameters
    ----------
    subject_score : float
        Alignment raw score computed by BLAST.
    query_score : float
        Raw score computed by BLAST for the
        self-alignment of the query sequence.

    Returns
    -------
    bsr : float
        BLAST Score Ratio for the alignment.
    """
    bsr = subject_score / query_score

    return bsr


def determine_self_scores(fasta_file, output_directory, makeblastdb_path,
                          blast_path, db_type, blast_threads):
    """Compute the self-alignment raw score for sequences in a FASTA file.

    Parameters
    ----------
    fasta_file : str
        Path to the FASTA file that contains the sequences.
    output_directory : str
        Path to the working directory.
    makeblastdb_path : str
        Path to the 'makeblastdb' executable.
    blast_path : str
        Path to the BLASTp/n executable.
    db_type : str
        Type of the database, nucleotide (nuc) or
        protein (prot).
    blast_threads : int
        Number of threads/cores used to run BLAST.

    Returns
    -------
    self_scores : dict
        Dictionary with sequence identifiers as keys and
        tuples with the sequence length and the raw score
        of the self-alignment as values.
    """
    # change identifiers to shorten and avoid BLAST error
    # related with sequence header length
    integer_fasta = fasta_file.replace('.fasta', '_integer_headers.fasta')
    ids_map = fao.integer_headers(fasta_file, integer_fasta,
                                  start=1, limit=5000, prefix='seq_')

    blast_db = fo.join_paths(output_directory,
                             [fo.file_basename(integer_fasta, False)])
    # will not work if file contains duplicates
    db_stderr = bw.make_blast_db(makeblastdb_path, integer_fasta,
                                 blast_db, db_type)

    # split Fasta file to BLAST short sequences (<30aa) separately
    # only possible to have alleles <30aa with non-default schemas
    above_outfile, below_outfile = fao.split_seqlength(integer_fasta,
                                                       output_directory,
                                                       ct.BLAST_TASK_THRESHOLD['blastp'])

    if above_outfile is not None:
        # divide FASTA file into groups of 100 sequences to reduce
        # execution time for large sequence sets
        splitted_fastas = fao.split_seqcount(above_outfile[0], output_directory, 100)

        # create TXT with list of sequence identifiers
        seqids_files = []
        for f in splitted_fastas:
            seqids = list(f[1])
            seqids_file = fo.join_paths(output_directory, [fo.file_basename(f[0], False)])
            fo.write_lines(seqids, seqids_file)
            seqids_files.append(seqids_file)
    # this should not happen or be very rare, but just in case
    else:
        splitted_fastas = []

    # create directory to store results from final BLASTp
    final_blastp_dir = fo.join_paths(output_directory, ['BLAST_results'])
    fo.create_directory(final_blastp_dir)
    blast_outputs = ['{0}/{1}_blast_out.tsv'.format(final_blastp_dir,
                                                    fo.file_basename(file[0], False))
                     for file in splitted_fastas]

    # add common arguments to all sublists
    blast_inputs = [[blast_path, blast_db, file[0],
                     blast_outputs[i], 1, 1,
                     seqids_files[i], 'blastp', None,
                     ct.IGNORE_RAISED, None, bw.run_blast]
                    for i, file in enumerate(splitted_fastas)]

    # add file with short sequences
    if below_outfile is not None:
        seqids = list(below_outfile[1])
        seqids_file = fo.join_paths(output_directory, [fo.file_basename(below_outfile[0], False)])
        fo.write_lines(seqids, seqids_file)
        below_blastout = '{0}/{1}_blast_out.tsv'.format(final_blastp_dir,
                                                        fo.file_basename(below_outfile[0], False))
        blast_outputs.append(below_blastout)
        blast_inputs.append([blast_path, blast_db, below_outfile[0],
                             below_blastout, 1, 1,
                             seqids_file, 'blastp-short', None,
                             ct.IGNORE_RAISED, None, bw.run_blast])

    blast_stderr = mo.map_async_parallelizer(blast_inputs,
                                             mo.function_helper,
                                             blast_threads,
                                             show_progress=False)

    blast_stderr = im.flatten_list(blast_stderr)
    if len(blast_stderr) > 0:
        sys.exit(blast_stderr)

    # concatenate files with BLASTp results
    blast_output = fo.join_paths(final_blastp_dir, ['blast_out_concat.tsv'])
    blast_output = fo.concatenate_files(blast_outputs, blast_output)

    current_results = fo.read_tabular(blast_output)
    # get raw score and sequence length
    # multiply by 3 to get DNA sequence length and add 3 to count stop codon
    self_results = [line for line in current_results if line[0] == line[4]]
    self_scores = {}
    for line in self_results:
        # multiply by 3 to get DNA sequence length
        # add 3 to count stop codon
        dna_length = (int(line[3])*3)+3
        self_scores[ids_map[line[0]]] = (dna_length, float(line[6]))

    # determine if we got the score for all representatives
    if len(self_scores) < len(ids_map):
        missing = [seqid
                   for seqid in ids_map
                   if seqid not in [line[0] for line in self_results]]
        # index FASTA file
        concat_reps_index = fao.index_fasta(integer_fasta)
        for seqid in missing:
            # get sequence
            current_rep = concat_reps_index[seqid]
            record = fao.fasta_str_record(ct.FASTA_RECORD_TEMPLATE, [current_rep.id, str(current_rep.seq)])
            rep_file = fo.join_paths(output_directory, [f'{current_rep.id}_solo.fasta'])
            fo.write_lines([record], rep_file)
            # save file with ids
            id_file = fo.join_paths(output_directory, [f'{current_rep.id}_ids.txt'])
            fo.write_lines([current_rep.id], id_file)
            rep_blastout = fo.join_paths(output_directory, [f'{current_rep.id}_blastout.tsv'])
            # disable composition-based stats
            bw.run_blast(blast_path, blast_db, rep_file,
                         rep_blastout, 1, 1,
                         id_file, 'blastp', None,
                         ct.IGNORE_RAISED, 0)
            rep_results = fo.read_tabular(rep_blastout)
            if len(rep_results) > 0:
                dna_length = (int(rep_results[0][3])*3)+3
                self_scores[ids_map[rep_results[0][0]]] = (dna_length, float(rep_results[0][6]))
            else:
                print('Could not determine the self-alignment raw '
                      f'score for {ids_map[rep_results[0][0]]}')

    return self_scores



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Purpose
-------

This module contains functions to work with FASTA files.

Code documentation
------------------
"""


from Bio import SeqIO
from Bio.SeqIO import FastaIO

try:
    from utils import (file_operations as fo,
                       iterables_manipulation as im,
                       sequence_manipulation as sm,
                       constants as ct)
except ModuleNotFoundError:
    from CHEWBBACA.utils import (file_operations as fo,
                                 iterables_manipulation as im,
                                 sequence_manipulation as sm,
                                 constants as ct)


def sequence_generator(input_file):
    """Create a SeqRecord iterator.

    Parameters
    ----------
    input_file : str
        Path to a Fasta file.

    Returns
    -------
    records : Bio.SeqIO.FastaIO.FastaIterator
        SeqRecord iterator.
    """
    records = SeqIO.parse(input_file, 'fasta')

    return records


def index_fasta(fasta_file):
    """Create index to retrieve data from a FASTA file.

    Parameters
    ----------
    fasta_file : str
        Path to a FASTA file.

    Returns
    -------
    fasta_index : Bio.File._IndexedSeqFileDict
        FASTA file index.
    """
    fasta_index = SeqIO.index(fasta_file, 'fasta')

    return fasta_index


def import_sequences(input_file):
    """Import sequences from a FASTA file.

    Parameters
    ----------
    input_file : str
        Path to a FASTA file.

    Returns
    -------
    seqs_dict : dict
        Dictionary with sequence identifiers as keys and
        sequences as values.
    """
    records = sequence_generator(input_file)
    seqs_dict = {rec.id: str(rec.seq.upper()) for rec in records}

    return seqs_dict


def count_sequences(fasta_file):
    """Count the number of sequences in a FASTA file.

    Parameters
    ----------
    fasta_file : str
        Path to a FASTA file.

    Returns
    -------
    total_seqs : int
        Number of sequences in the FASTA file.
    """
    with open(fasta_file, 'r') as infile:
        sequence_headers = [line for line in infile if line.startswith('>')]
        total_seqs = len(sequence_headers)

    return total_seqs


def write_records(records, output_file):
    """Write FASTA records (BioPython SeqRecord) to a file.

    Parameters
    ----------
    records : list
        List with BioPython SeqRecord objects.
    output_file : str
        Path to the output file.
    """
    with open(output_file, 'w') as output_handle:
        fasta_out = FastaIO.FastaWriter(output_handle, wrap=None)
        fasta_out.write_file(records)


def integer_headers(input_fasta, output_fasta, start=1,
                    limit=50000, prefix='', id_map=True):
    """Switch sequence headers in Fasta file by integer values.

    Parameters
    ----------
    input_fasta : str
        Path to a FASTA file.
    output_fasta : str
        Path to the output file with modified headers.
    start : int
        Integer value for the first identifier.
    limit : int
        Maximum number of FASTA records to keep in
        memory.
    prefix : str
        Prefix added to the sequence identifiers.
    id_map : bool
        If True, the function returns a dictionary
        mapping the original sequence IDs to the
        new sequence IDs. If False, returns the
        number of sequences renamed.

    Returns
    -------
    ids_map : dict
        Dictionary with mapping between integer and original
        headers.
    """
    seqs = []
    ids_map = {}
    exausted = False
    seq_generator = sequence_generator(input_fasta)
    while exausted is False:
        record = next(seq_generator, None)
        if record is not None:
            # new_id = 'seq_{0}'.format(start)
            new_id = '{0}{1}'.format(prefix, start)
            ids_map[new_id] = record.id
            sequence = str(record.seq)
            new_rec = fasta_str_record(ct.FASTA_RECORD_TEMPLATE,
                                       [new_id, sequence])
            seqs.append(new_rec)
            start += 1
        elif record is None:
            exausted = True

        if len(seqs) == limit or exausted is True:
            fo.write_lines(seqs, output_fasta, write_mode='a')
            seqs = []

    if id_map is True:
        return ids_map
    else:
        return len(ids_map)


def fasta_str_record(record_template, record_data):
    """Create the string representation of a FASTA record.

    Parameters
    ----------
    record_template : str
        String template to construct the FASTA record.
    record_data : list
        List with the elements to add to the string.

    Returns
    -------
    record : str
        String representation of the FASTA record.
    """
    record = record_template.format(*record_data)

    return record


def fasta_lines(template, records_data):
    """Create a list with FASTA records.

    Parameters
    ----------
    template : str
        String template to construct the FASTA record.
    records_data : list
        A list with one sublist per FASTA record.
        Each sublist contains the elements to insert
        inside the template placeholders.

    Returns
    -------
    seqs_lines : list
        A list with strings representing FASTA records.
    """
    seqs_lines = [fasta_str_record(template, arg) for arg in records_data]

    return seqs_lines


def validate_fasta(file_path):
    """Check if a file is a FASTA file.

    Parameters
    ----------
    file_path : str
        Path to the file.

    Returns
    -------
    True if file has valid FASTA format,
    False otherwise.
    """
    records = SeqIO.parse(file_path, 'fasta')

    # returns False if it was not a FASTA file
    return any(records)


def filter_non_fasta(files):
    """Select FASTA files from a list with file paths.

    Parameters
    ----------
    files : list
        A list that contains file paths.

    Returns
    -------
    fasta_files : list
        List that contains paths to FASTA files.
    """
    fasta_files = [file for file in files if validate_fasta(file) is True]

    return fasta_files


def sequence_lengths(fasta_file, hashed=False):
    """Determine length of sequences in a FASTA file.

    Read Fasta file and create dictionary with mapping
    between sequence identifiers and sequence lengths.

    Parameters
    ----------
    fasta_file : str
        Path to a FASTA file.
    hashed : bool
        If False, sequence identifiers are extracted from
        sequence headers. If True, sequence hashes will be
        used as keys.

    Returns
    -------
    lengths : dict
        Dictionary with sequence identifiers as keys and
        sequence lengths as values.
    """
    records = sequence_generator(fasta_file)
    if hashed is False:
        lengths = {rec.id: len(rec.seq) for rec in records}
    else:
        lengths = {im.hash_sequence(str(rec.seq)):
                   len(rec.seq) for rec in records}

    return lengths


def get_sequences_by_id(sequences, seqids, output_file, limit=50000):
    """Retrieve sequences from indexed FASTA file or dictionary.

    Retrieves sequences based on sequence identifiers from a FASTA
    file indexed with SeqIO.index or from a dictionary with sequence
    identifiers as keys and sequences as values.

    Parameters
    ----------
    sequences : dict or Bio.File._IndexedSeqFileDict
        Dictionary with seqids as keys and sequences as values or
        a FASTA file index created with BioPython.
    seqids : list
        List with the identifiers of the sequences that should be
        retrieved.
    output_file : str
        Path to the FASTA file to which selected sequences will
        be saved.
    limit : int
        Maximum number of sequences that will be kept in memory
        at a time (to avoid keeping huge datasets in memory).

    Returns
    -------
    total_selected : int
        Total number of records written to the output file.
    """
    # using a generator
    if type(sequences) == dict:
        seqs = ((seqid, sequences[seqid]) for seqid in seqids)
    else:
        seqs = ((seqid, str(sequences[seqid].seq)) for seqid in seqids)

    records = []
    total_selected = 0
    exhausted = False
    while exhausted is False:
        record = next(seqs, None)
        if record is not None:
            record = fasta_str_record(ct.FASTA_RECORD_TEMPLATE,
                                      [record[0], record[1]])
            records.append(record)
        else:
            exhausted = True

        # write records when it reaches the maximum number of records to
        # keep in memory or there are no records left to fetch
        if len(records) == limit or exhausted is True:
            fo.write_lines(records, output_file, write_mode='a')
            total_selected += len(records)
            records = []

    return total_selected


def split_seqcount(fasta_path, output_directory, max_seqs):
    """Split a FASTA file based on a maximum number of sequences per file.

    Parameters
    ----------
    fasta_path : str
        Path to a FASTA file.
    output_directory : str
        Path to the output directory.
    max_seqs : int
        Split FASTA file into files with a maximum number
        of sequences equal to this value.

    Returns
    -------
    splitted_files : list
        List with paths to the new files that were
        created by splitting the input FASTA file.
    """
    file_count = 1
    exhausted = False
    current_recs = []
    splitted_files = []
    record_generator = sequence_generator(fasta_path)
    while exhausted is False:
        record = next(record_generator, None)
        if record is not None:
            current_recs.append(record)
        else:
            exhausted = True

        if len(current_recs) == max_seqs or exhausted is True:
            if len(current_recs) > 0:
                file_path = fo.join_paths(output_directory,
                                          ['seqcount{0}.fasta'.format(file_count)])
                seqids = (rec.id for rec in current_recs)
                splitted_files.append([file_path, seqids])
                write_records(current_recs, file_path)
                current_recs = []
                file_count += 1

    return splitted_files


def split_seqlength(fasta_path, output_directory, length_cutoff):
    """Split a FASTA file based on a sequence length threshold.

    Parameters
    ----------
    fasta_path : str
        Path to a FASTA file.
    output_directory : str
        Path to the output directory.
    length_cutoff : int
        Sequence length threshold used to split the FASTA file.

    Returns
    -------
    List with two tuples: the first contains the path to a FASTA
    file with sequences equal or above the length cutoff and the
    list of sequence identifiers of those sequences or is None if
    there were no sequences above the cutoff. The second tuple
    contains the same data for the sequences below the cutoff or
    is None if there were no sequences below the cutoff.
    """
    length_values = sequence_lengths(fasta_path)
    below_cutoff = [seqid for seqid, length in length_values.items()
                    if length < length_cutoff]
    above_cutoff = list(set(length_values) - set(below_cutoff))

    fasta_index = index_fasta(fasta_path)
    if len(above_cutoff) > 0:
        above_outfile = fo.join_paths(output_directory, ['above_cutoff.fasta'])
        above_count = get_sequences_by_id(fasta_index, above_cutoff, above_outfile)
        above_seqids = (seqid for seqid in above_cutoff)
        above_data = [above_outfile, above_seqids]
    else:
        above_data = None

    # file has sequences shorter than cutoff value
    if len(below_cutoff) > 0:
        below_outfile = fo.join_paths(output_directory, ['below_cutoff.fasta'])
        below_count = get_sequences_by_id(fasta_index, below_cutoff, below_outfile)
        below_seqids = (seqid for seqid in below_cutoff)
        below_data = [below_outfile, below_seqids]
    else:
        below_data = None

    return [above_data, below_data]


def fasta_stats(fasta_file):
    """Determine the number of sequences in a FASTA file and length stats.

    Parameters
    ----------
    fasta_file : str
        Path to a FASTA file.

    Returns
    -------
    fasta_file : str
        Path to the FASTA file.
    total_seqs: int
        Total number of records in the FASTA file.
    mean_length: float
        Mean sequence length.
    """
    seq_lengths = sequence_lengths(fasta_file)
    min_length = min(seq_lengths.values())
    max_length = max(seq_lengths.values())
    mean_length = sum(seq_lengths.values())/len(seq_lengths)
    total_seqs = len(seq_lengths)

    return [fasta_file, total_seqs, min_length, max_length, mean_length]


def translate_fasta(input_fasta, output_directory, translation_table):
    """Translate DNA sequences in a FASTA file.

    Parameters
    ----------
    input_fasta : str
        Path to the FASTA file that contains the DNA sequences
        to translate.
    output_directory : str
        Path to the output directory where the FASTA file with
        protein sequences will be written to.
    translation_table : int
        Genetic code used to translate DNA sequences.

    Returns
    -------
    input_fasta : str
        Path to the input FASTA file.
    protein_file : str
        Path to the FASTA file that contains the translated
        sequences.
    Also returns the number of sequences that were translated
    successfully.
    """
    records = sequence_generator(input_fasta)
    translated_records = [[rec.id,
                           str(sm.translate_dna(str(rec.seq),
                                                translation_table, 0)[0][0])]
                          for rec in records]
    translated_lines = fasta_lines(ct.FASTA_RECORD_TEMPLATE,
                                   translated_records)

    basename = fo.file_basename(input_fasta, True).replace('.fasta', '_protein.fasta')
    protein_file = fo.join_paths(output_directory, [basename])

    fo.write_lines(translated_lines, protein_file)

    return [input_fasta, protein_file, len(translated_records)]
